from dtcontrol.decision_tree.determinization.determinizer import Determinizer

class NonDeterminizer(Determinizer):
    """
    This determinizer doesn't actually do any determinization but simply uses the label powerset approach.
    """

    def determinize(self, dataset):
        if self.pre_determinized_labels is not None:
            return self.pre_determinized_labels[dataset.parent_mask]
        return dataset.get_unique_labels()

    def is_pre_split(self):
        return True



from abc import ABC, abstractmethod

import numpy as np

from dtcontrol.dataset.csv_dataset_loader import CSVDatasetLoader
from dtcontrol.dataset.prism_dataset_loader import PrismDatasetLoader
from dtcontrol.dataset.scots_dataset_loader import ScotsDatasetLoader
from dtcontrol.dataset.uppaal_dataset_loader import UppaalDatasetLoader
from dtcontrol.util import get_filename_and_ext

class Dataset(ABC):
    """
    Attributes:
        x: the training data of shape (num_states, num_state_dims)
        y:
            Multi-output: the training labels of shape (num_input_dims, num_states, max_non_determinism)
                For example:  [ [[1  2 -1]  [[4  5 -1]
                                [2  3 -1]   [4  6 -1]
                                [1  2  3]   [4  6  6]
                                [1 -1 -1]]  [5 -1 -1]] ]
                has the shape (2, 4, 3) as there are two control inputs, 4 states and
                at most 3 non-deterministic choices for each state.

                The Y_train in the above example can be thought of a list [y1, y2]
                where y1 gives the values for the first control input and y2 gives
                the values for the second control input. Let us for example see
                which inputs are allowed for the 2nd state.

                y1[1] = [2  3 -1] and y2[1] = [4  6 -1]

                This means that the allowed control inputs for the 2nd state are
                (2, 4) and (3, 6). The -1 is a filler just to make the length of
                the lists  = max_non_determinism.

                Use np.stack(Y_train, axis=2) in order to get an array of the form
                [[[1  4], [ 2  5], [-1 -1]],
                 [[2  4], [ 3  6], [-1 -1]]
                 [[1  4], [ 2  6], [ 3  6]]
                 [[1  5], [-1 -1], [-1 -1]]

                from which it is easier to extract control actions as tuples.

            Single-output: the training labels of shape (num_states, max_non_determinism)
                For example:   [[1  2 -1]
                                [2  3 -1]
                                [1  2  3]
                                [1 -1 -1]]
                has the shape (4, 3) as there are 4 states and at most 3 non-deterministic
                choices for each state.

                y[i] gives the allowed control inputs for the ith state. -1 is a
                filler just to make the length of the lists  = max_non_determinism.
    """

    def __init__(self, filename):
        self.filename = filename
        self.name, self.extension = get_filename_and_ext(filename)
        self.extension_to_loader = {
            '.scs': ScotsDatasetLoader(),
            '.dump': UppaalDatasetLoader(),
            '.csv': CSVDatasetLoader(),
            '.prism': PrismDatasetLoader()
        }
        if self.extension not in self.extension_to_loader:
            raise ValueError('Unknown file format.')
        self.x = None
        self.numeric_x = None
        self.categorical_x = None
        self.x_metadata = {"variables": None, "categorical": None, "category_names": None,
                           "min": None, "max": None, "step_size": None}
        self.y = None
        self.y_metadata = {"categorical": None, "category_names": None, "min": None, "max": None, "step_size": None,
                           'num_rows': None, 'num_flattened': None}
        self.index_to_actual = {}  # mapping from arbitrary integer indices to the actual float/categorical labels
        self.numeric_feature_mapping = {}  # maps indices in the numeric array to the actual column index in x
        self.numeric_columns = None
        self.categorical_feature_mapping = {}  # the same thing for the categorical array
        self.categorical_columns = None
        self.is_deterministic = None
        self.parent_mask = None  # if this is a subset, parent_mask saves the mask into the parent dataset

    def get_name(self):
        return self.name

    def copy_from_other_dataset(self, ds):
        self.x = ds.x
        self.numeric_x = None
        self.categorical_x = None
        self.x_metadata = ds.x_metadata
        self.y = ds.y
        self.y_metadata = ds.y_metadata
        self.index_to_actual = ds.index_to_actual
        self.numeric_feature_mapping = ds.numeric_feature_mapping
        self.categorical_feature_mapping = ds.categorical_feature_mapping
        self.is_deterministic = ds.is_deterministic

    def load_if_necessary(self):
        if self.x is None:
            self.x, self.x_metadata, self.y, self.y_metadata, self.index_to_actual = \
                self.extension_to_loader[self.extension].load_dataset(self.filename)
            assert len([i for i in self.index_to_actual if i == 0]) == 0  # labels have to start with 1 because of OC1
            self.y_metadata['num_rows'] = len(self.x)
            self.y_metadata['num_flattened'] = sum(1 for row in self.y for y in row)

    def load_metadata_from_json(self, json_object):
        metadata = json_object['metadata']
        self.x_metadata = metadata['X_metadata']
        self.y_metadata = metadata['Y_metadata']

    def check_loaded(self):
        if self.x is None:
            raise RuntimeError('Dataset is not loaded.')

    def get_numeric_x(self):
        if self.numeric_x is None:
            self.numeric_columns = set(range(self.x.shape[1])).difference(set(self.x_metadata['categorical']))
            self.numeric_columns = sorted(list(self.numeric_columns))
            self.numeric_feature_mapping = {i: self.numeric_columns[i] for i in range(len(self.numeric_columns))}
            self.numeric_x = self.x[:, self.numeric_columns]
        return self.numeric_x

    def get_categorical_x(self):
        if self.categorical_x is None:
            self.categorical_columns = self.x_metadata['categorical']
            self.categorical_feature_mapping = {i: self.categorical_columns[i] for i in
                                                range(len(self.categorical_columns))}
            self.categorical_x = self.x[:, self.categorical_columns]
        return self.categorical_x

    def map_numeric_feature_back(self, feature):
        return self.numeric_feature_mapping[feature]

    def map_categorical_feature_back(self, feature):
        return self.categorical_feature_mapping[feature]

    def __len__(self):
        return len(self.x)

    @abstractmethod
    def compute_accuracy(self, y_pred):
        pass

    @abstractmethod
    def get_single_labels(self):
        """
        Converts multi-output labels to the tuple-id-representation, resulting in single-output labels that are
        returned.
        For a SingleOutputDataset, simply returns y.
        """
        pass

    @abstractmethod
    def map_single_label_back(self, single_label):
        """
        For a multi-output dataset, returns the tuple corresponding to the tuple id.
        For a single-output dataset, simply returns the single-label.
        """
        pass

    @abstractmethod
    def get_unique_labels(self):
        """
        Returns a label representation in which each combination of possible labels occurring in the data is
        assigned a new unique label.
        """
        pass

    def index_label_to_actual(self, index_label):
        """
        :param index_label: the index label
        :returns: the actual (float/categorical) label, which can either be a single label, a single tuple, a list of
                  labels, or a list of tuples
        """
        if isinstance(index_label, tuple):  # single tuple
            return tuple([self.index_to_actual[i] for i in index_label if i != -1])
        elif isinstance(index_label, list):
            if isinstance(index_label[0], tuple):  # list of tuples
                return [tuple(map(lambda x: self.index_to_actual[x], tup)) for tup in index_label]
            else:  # list of labels
                return [self.index_to_actual[i] for i in index_label if i != -1]
        else:  # single label
            return self.index_to_actual[index_label]

    @abstractmethod
    def map_unique_label_back(self, label):
        """
        :param label: the unique label
        :return: the corresponding index (int) label
        """
        pass

    @abstractmethod
    def from_mask(self, mask):
        """
        Returns the subset given by the mask.
        :param mask: a numpy array of 0s and 1s with len(mask) == num_examples
        """
        pass

    @abstractmethod
    def from_mask_optimized(self, mask):
        """
        An optimized version of the from_mask method that only copies the labels. This is used inside the critical
        impurity measure code.
        """
        pass

    @staticmethod
    def get_unique_labels_from_2d(labels):
        """
        Computes unique labels of a 2d label array by mapping every unique inner array to an int. Returns the unique labels
        and the int mapping.
        """
        l = []
        int_to_label = {}
        next_unused_int = 1  # OC1 expects labels starting with 1
        label_str_to_int = {}
        for i in range(len(labels)):
            label_str = ','.join(sorted([str(i) for i in labels[i] if i != -1]))
            if label_str not in label_str_to_int:
                label_str_to_int[label_str] = next_unused_int
                int_to_label[next_unused_int] = labels[i]
                next_unused_int += 1
            new_label = label_str_to_int[label_str]
            l.append(new_label)
        return np.array(l), int_to_label



import logging
from os.path import basename, splitext

import numpy as np

def ignore_convergence_warnings():
    logging.captureWarnings(capture=True)
    logger = logging.getLogger("py.warnings")
    handler = logging.StreamHandler()
    logger.addHandler(handler)
    logger.addFilter(lambda record: "ConvergenceWarning" not in record.getMessage())

def format_seconds(sec):
    m, s = divmod(sec, 60)
    h, m = divmod(m, 60)
    d, h = divmod(h, 24)
    pattern = '%%02d:%%02d:%%0%d.%df' % (6, 3)
    if d == 0:
        return pattern % (h, m, s)
    return ('%d days, ' + pattern) % (d, h, m, s)

def get_filename_and_ext(filename):
    path, ext = splitext(filename)
    return basename(path), ext

def make_set(v):
    if v is None:
        return set()
    if isinstance(v, tuple):
        return {v}
    try:
        return set(v)
    except TypeError:
        return {v}

def objround(obj, precision):
    if isinstance(obj, list) or isinstance(obj, np.ndarray):
        return [objround(o, precision) for o in obj]
    if isinstance(obj, tuple):
        return tuple(round(o, precision) for o in obj)
    # if just a float
    return round(obj, precision)

def print_tuple(t):
    return f'({", ".join([str(e) for e in t])})'

def print_list(l):
    return f'[{", ".join([str(e) for e in l])}]'

def print_set(s):
    return f'{{{", ".join([str(e) for e in s])}}}'

def split_into_lines(l):
    if not isinstance(l, list) or len(l) < 5:
        return print_list(l)
    i = 0
    l2 = []
    while i < len(l):
        l2.append(', '.join([str(j) for j in l[i:min(i + 5, len(l))]]))
        i += 5
    return '[' + '\\n'.join(l2) + ']'

def peek_line(file):
    pos = file.tell()
    line = file.readline()
    file.seek(pos)
    return line

def log_without_newline(msg):
    old_terminator = logging.StreamHandler.terminator
    logging.StreamHandler.terminator = ""
    logging.info(msg)
    logging.StreamHandler.terminator = old_terminator

def is_int(s):
    try:
        int(s)
        return True
    except ValueError:
        return False



import sys

import numpy as np

from dtcontrol.decision_tree.impurity.deterministic_impurity_measure import DeterministicImpurityMeasure

class TwoingRule(DeterministicImpurityMeasure):
    def calculate_impurity(self, dataset, split):
        if len(split.get_masks(dataset)) == 1:
            return sys.maxsize
        assert len(split.get_masks(dataset)) == 2
        [left_mask, right_mask] = split.get_masks(dataset)
        left = self.determinizer.determinize(dataset.from_mask_optimized(left_mask))
        right = self.determinizer.determinize(dataset.from_mask_optimized(right_mask))
        if len(left) == 0 or len(right) == 0:
            return sys.maxsize
        num_labels = len(dataset)
        twoing_value = (len(left) / num_labels) * (len(right) / num_labels)
        s = 0
        unique = np.unique(y)
        for u in unique:
            num_left = len(left[left == u])
            num_right = len(right[right == u])
            s += abs(num_left / len(left) - num_right / len(right))
        twoing_value *= s * s
        if twoing_value == 0:
            return sys.maxsize
        return 1 / twoing_value

    def get_oc1_name(self):
        return 'twoing'



import sys

import numpy as np
from sklearn.linear_model import LogisticRegression
from sklearn.metrics import roc_auc_score

from dtcontrol.decision_tree.impurity.deterministic_impurity_measure import DeterministicImpurityMeasure

class AUROC(DeterministicImpurityMeasure):
    def calculate_impurity(self, dataset, split):
        if len(split.get_masks(dataset)) == 1:
            return sys.maxsize
        scores = []
        for mask in split.get_masks(dataset):
            subset_labels = self.determinizer.determinize(dataset.from_mask_optimized(mask))
            scores.append(self.calculate_auroc(dataset.x[mask], subset_labels))
        if any([s == 0 for s in scores]):
            return sys.maxsize
        return 1 / sum(scores)

    @staticmethod
    def calculate_auroc(x, y):
        if len(y) == 0:
            return 0
        unique_labels = np.unique(y)
        if len(unique_labels) == 1:
            return 1  # perfect split
        label_to_auroc = {}
        for label in unique_labels:
            new_y = np.copy(y)
            label_mask = (new_y == label)
            new_y[label_mask] = 1
            new_y[~label_mask] = -1
            logreg = LogisticRegression(solver='lbfgs', penalty='none')
            logreg.fit(x, new_y)
            auroc = roc_auc_score(new_y, logreg.decision_function(x))
            label_to_auroc[label] = auroc
        weighted_avg = sum([auroc * (len(y[y == label]) / len(y)) for label, auroc in label_to_auroc.items()])
        return weighted_avg

    def get_oc1_name(self):
        return None



import numpy as np

from dtcontrol.decision_tree.determinization.determinizer import Determinizer

class MaxFreqDeterminizer(Determinizer):
    """
    This determinizer uses the maximum frequency determinization approach.
    """

    def __init__(self, pre_determinize=True):
        super().__init__()
        self.pre_determinize = pre_determinize

    def determinize(self, dataset):
        if self.is_pre_split() and self.pre_determinized_labels is not None:
            return self.pre_determinized_labels[dataset.parent_mask]
        return self.get_max_freq_labels(dataset.get_single_labels())

    @staticmethod
    def get_max_freq_labels(labels):
        flattened_labels = labels.flatten()
        # remove -1 as we use it only as a filler
        flattened_labels = flattened_labels[flattened_labels != -1]
        label_counts = np.bincount(flattened_labels)
        new_labels = []
        for i in range(len(labels)):
            current = labels[i]
            current = current[current != -1]
            max_label = max(list(current), key=lambda l: label_counts[l])
            assert max_label != -1
            new_labels.append(max_label)
        return np.array(new_labels)

    def is_pre_split(self):
        return self.pre_determinize



#!/usr/bin/env python

"""
README

Run dtcontrol --help to see usage.
"""

import argparse
import logging
import re
import shutil
import sys
from collections import namedtuple, OrderedDict
from os import makedirs, remove
from os.path import exists, isfile, splitext
from typing import Tuple, Union, List

import pkg_resources
from pkg_resources import Requirement, resource_filename
from ruamel.yaml import YAML
from ruamel.yaml.scanner import ScannerError
from sklearn.linear_model import LogisticRegression
from sklearn.svm import LinearSVC
from tabulate import tabulate

from dtcontrol.benchmark_suite import BenchmarkSuite
from dtcontrol.decision_tree.decision_tree import DecisionTree
# Import determinizers
from dtcontrol.decision_tree.determinization.max_freq_determinizer import MaxFreqDeterminizer
from dtcontrol.decision_tree.determinization.non_determinizer import NonDeterminizer

# Import impurity measures
from dtcontrol.decision_tree.impurity.auroc import AUROC
from dtcontrol.decision_tree.impurity.entropy import Entropy
from dtcontrol.decision_tree.impurity.gini_index import GiniIndex
from dtcontrol.decision_tree.impurity.max_minority import MaxMinority
from dtcontrol.decision_tree.impurity.multi_label_entropy import MultiLabelEntropy
from dtcontrol.decision_tree.impurity.multi_label_gini_index import MultiLabelGiniIndex
from dtcontrol.decision_tree.impurity.multi_label_twoing_rule import MultiLabelTwoingRule
from dtcontrol.decision_tree.impurity.twoing_rule import TwoingRule

# Import preprocessing strategies
from dtcontrol.decision_tree.pre_processing.norm_pre_processor import NormPreProcessor
from dtcontrol.decision_tree.pre_processing.random_pre_processor import RandomPreProcessor

# Import splitting strategies
from dtcontrol.decision_tree.splitting.axis_aligned import AxisAlignedSplittingStrategy
from dtcontrol.decision_tree.splitting.categorical_multi import CategoricalMultiSplittingStrategy
from dtcontrol.decision_tree.splitting.categorical_single import CategoricalSingleSplittingStrategy
from dtcontrol.decision_tree.splitting.linear_classifier import LinearClassifierSplittingStrategy
from dtcontrol.decision_tree.splitting.oc1 import OC1SplittingStrategy  # TODO Doesn't work on MacOS, check
from dtcontrol.post_processing.safe_pruning import SafePruning


def main():
    def is_valid_file_or_folder(parser, arg):
        if not exists(arg):
            parser.error(f"The file/folder {arg} does not exist.")
        else:
            return arg

    def is_valid_file(parser, arg):
        if not isfile(arg):
            parser.error(f"The file {arg} does not exist. Give a valid JSON file path.")
        else:
            return arg

    def load_default_config() -> OrderedDict:
        try:
            default_config_file = resource_filename(Requirement.parse("dtcontrol"),
                                                    "config.yml")  # System-level config file
        except pkg_resources.DistributionNotFound:
            sys.exit(
                f"pkg_resources could not find a distribution called 'dtcontrol'. Please report this error to the developers.")

        try:
            yaml = YAML()
            default_config = yaml.load(open(default_config_file))
        except FileNotFoundError:
            sys.exit(f"Error finding the default config file. Please raise an issue with the developers.")
        except ScannerError:
            sys.exit(
                f"Scan error in the default YAML configuration file '{default_config_file}'. Please raise an issue with the developers.")
        return default_config

    def parse_timeout(timeout_string: str) -> int:
        """
        Parses the timeout string

        :param timeout_string: string describing timeout - an integer suffixed with s, m or h
        :return: timeout in seconds
        """
        # Default timeout set to 2 hours
        unit_to_factor = {'s': 1, 'm': 60, 'h': 3600}
        if re.match(r'^[0-9]+[smh]$', timeout_string):
            factor = unit_to_factor[timeout_string[-1]]
            timeout = int(args.timeout[:-1]) * factor
        else:
            # In case s, m or h is missing; then interpret number as timeout in seconds
            try:
                timeout = int(timeout_string)
            except ValueError:
                parser.error("Invalid value passed as timeout.")
        return timeout

    def preset_parser(args):
        if args.list:
            system_config = load_default_config()
            user_config = None

            if args.config:
                try:
                    yaml = YAML()
                    user_config = yaml.load(open(args.config))
                except FileNotFoundError:
                    sys.exit(
                        f"Error finding the config file. Please check if the file '{args.config}' exists in the current directory")
                except ScannerError:
                    sys.exit(f"Scan error in the YAML configuration file '{args.config}'. Please re-check syntax.")
                else:
                    if 'presets' not in user_config:
                        logging.warning("WARNING: config file does not contain the 'presets' key. "
                                        "No user configurations could be loaded. Ensure that user presets "
                                        "are defined under the presets key.\n")
            else:
                logging.info("WARNING: --config switch not used, only loading pre-defined run configurations.")

            list_presets(user_config, system_config)
        elif args.sample:
            print("# Paste the following contents into a .yml file and \n"
                  "# pass the file to dtcontrol using --config <filename>.yml\n"
                  "presets:\n"
                  "  my-config:\n"
                  "    determinize: maxfreq\n"
                  "    numeric-predicates: ['axisonly']\n"
                  "    categorical-predicates: ['singlesplit']\n"
                  "    impurity: 'entropy'\n"
                  "    safe-pruning: False\n"
                  "  another-config:\n"
                  "    determinize: minnorm\n"
                  "    numeric-predicates: ['linear-logreg']\n"
                  "    categorical-predicates: ['valuegrouping']\n"
                  "    tolerance: 10e-4\n"
                  "    safe-pruning: False")
        else:
            parser_conf.print_help()

        sys.exit(0)

    def get_key_or_default(preset_dict: OrderedDict, default_dict: OrderedDict, key: str):
        if key in preset_dict:
            return preset_dict[key]
        else:
            return default_dict[key]

    def get_key_or_empty(preset_dict: OrderedDict, default_dict: OrderedDict, key: str):
        if key in preset_dict:
            return preset_dict[key]
        else:
            return ""

    def list_presets(user_config: OrderedDict, system_config: OrderedDict):
        Row = namedtuple('Row',
                         ['Name', 'NumericPredicate', 'CategoricalPredicate', 'Determinize', 'Impurity', 'Tolerance',
                          'SafePruning'])

        run_config_table = []
        default_value: OrderedDict = load_default_config()['presets']['default']

        if user_config and 'presets' in user_config:
            for preset in user_config['presets']:
                value = user_config['presets'][preset]
                run_config_table.append(Row(Name=preset,
                                            NumericPredicate=get_key_or_default(value, default_value, 'numeric-predicates'),
                                            CategoricalPredicate=get_key_or_empty(value, default_value,
                                                                                  'categorical-predicates'),
                                            Determinize=get_key_or_default(value, default_value, 'determinize'),
                                            Impurity=get_key_or_default(value, default_value, 'impurity'),
                                            Tolerance=get_key_or_empty(value, default_value, 'tolerance'),
                                            SafePruning=get_key_or_empty(value, default_value, 'safe-pruning')))

        if run_config_table:
            logging.info("The following user presets (run configurations) are available:\n")
            logging.info(tabulate(run_config_table,
                                  ['name', 'numeric-predicates', 'categorical-predicates', 'determinize', 'impurity',
                                   'tolerance', 'safe-pruning'],
                                  tablefmt="presto"))
            logging.info("\n")

        run_config_table = []
        if system_config and 'presets' in system_config:
            for preset in system_config['presets']:
                value = system_config['presets'][preset]
                run_config_table.append(Row(Name=preset,
                                            NumericPredicate=get_key_or_default(value, default_value,
                                                                                'numeric-predicates'),
                                            CategoricalPredicate=get_key_or_empty(value, default_value,
                                                                                  'categorical-predicates'),
                                            Determinize=get_key_or_default(value, default_value, 'determinize'),
                                            Impurity=get_key_or_default(value, default_value, 'impurity'),
                                            Tolerance=get_key_or_empty(value, default_value, 'tolerance'),
                                            SafePruning=get_key_or_empty(value, default_value, 'safe-pruning')))

        logging.info("The following pre-defined presets (run configurations) are available:\n")
        logging.info(tabulate(run_config_table,
                              ['name', 'numeric-predicates', 'categorical-predicates', 'determinize', 'impurity',
                               'tolerance', 'safe-pruning'],
                              tablefmt="presto"))
        logging.info("\n")
        logging.info("User presets take precedence over pre-defined presets. "
                     "Try running, for example,\n\tdtcontrol --input examples/cartpole.scs --use-preset maxfreq")

    def fetch_presets_from_loaded_config(loaded_config: Union[None, OrderedDict]) -> List[str]:
        if loaded_config:
            return list(loaded_config['presets'].keys())
        else:
            return []

    def get_preset(preset: str, user_config: OrderedDict, default_config: OrderedDict) -> Tuple:
        if user_config and preset in user_config['presets']:
            value = user_config['presets'][preset]
        elif preset in default_config['presets']:
            value = default_config['presets'][preset]
        else:
            sys.exit(f"Preset '{preset}' not found.\n"
                     f"Please ensure that the specified config file contains a "
                     f"configuration called '{preset}'. You may run the command\n"
                     f"\tdtcontrol preset --sample > user-config.yml\n"
                     f"to generate a configuration file and use it with the help of the "
                     f"--config and --use-preset switches (see help).\n"
                     f"Refer to the User Manual (https://dtcontrol.readthedocs.io/) for details on how to write presets.")

        default_value = default_config['presets']['default']

        # Obtain the different settings from the value dict
        # In case something is not defined, choose the default from default_value dict
        for key in value.keys():
            if key not in ['numeric-predicates', 'categorical-predicates', 'determinize', 'impurity', 'tolerance', 'safe-pruning']:
                logging.warning(f"Ignoring unknown key {key} specified under preset {preset}.")

        numeric_predicates = get_key_or_default(value, default_value, 'numeric-predicates')
        categorical_predicates = get_key_or_default(value, default_value, 'categorical-predicates')
        determinize = get_key_or_default(value, default_value, 'determinize')
        impurity = get_key_or_default(value, default_value, 'impurity')
        tolerance = get_key_or_default(value, default_value, 'tolerance')
        safe_pruning = get_key_or_default(value, default_value, 'safe-pruning')

        return numeric_predicates, categorical_predicates, determinize, impurity, tolerance, safe_pruning

    # TODO Make early stopping user definable
    def get_classifier(numeric_split, categorical_split, determinize, impurity, tolerance=1e-5, safe_pruning=False,
                       name=None):
        """
        Creates classifier objects for each method

        :param name:
        :param safe_pruning:
        :param impurity:
        :param determinize:
        :param split:
        :param tolerance:
        :param value_grouping:
        :return: list of classifier objects
        """

        combined_split = numeric_split + categorical_split
        # Give the preset a name, if doesn't exist
        if not name:
            name = f"{determinize}-({','.join(combined_split)})-{impurity}"

        if not isinstance(tolerance, float):
            raise ValueError(f"{tolerance} is not a valid tolerance value (enter a float, e.g., 1e-5). Exiting...")

        if not isinstance(safe_pruning, bool):
            raise ValueError(f"{safe_pruning} is not a valid value for safe-pruning in preset {name}. Exiting...")

        determinization_map = {
            'maxfreq': lambda x: MaxFreqDeterminizer(),
            'minnorm': lambda x: NormPreProcessor(min),
            'maxnorm': lambda x: NormPreProcessor(max),
            'random': lambda x: RandomPreProcessor(),
            'none': lambda x: NonDeterminizer(),
            'auto': lambda x: MaxFreqDeterminizer()
        }
        splitting_map = {
            'axisonly': lambda x: AxisAlignedSplittingStrategy(),
            'linear-logreg': lambda x: LinearClassifierSplittingStrategy(LogisticRegression, determinizer=x, solver='lbfgs', penalty='none'),
            'linear-linsvm': lambda x: LinearClassifierSplittingStrategy(LinearSVC, determinizer=x, max_iter=5000),
            'oc1': lambda x: OC1SplittingStrategy(determinizer=x),  # TODO See import comment, doesn't work on Mac
            'multisplit': lambda x: CategoricalMultiSplittingStrategy(value_grouping=False),
            'singlesplit': lambda x: CategoricalSingleSplittingStrategy(),
            'valuegrouping': lambda x: CategoricalMultiSplittingStrategy(value_grouping=True, tolerance=tolerance),
        }
        impurity_map = {
            'auroc': lambda x: AUROC(determinizer=x),
            'entropy': lambda x: Entropy(determinizer=x),
            'gini': lambda x: GiniIndex(determinizer=x),
            'maxminority': lambda x: MaxMinority(determinizer=x),
            'twoing': lambda x: TwoingRule(determinizer=x),
            'multilabelentropy': lambda x: MultiLabelEntropy(),
            'multilabelgini': lambda x: MultiLabelGiniIndex(),
            'multilabeltwoing': lambda x: MultiLabelTwoingRule(),
        }

        # Sanity check
        for sp in combined_split:
            if sp not in splitting_map:
                raise ValueError(f"{sp} is not a valid predicate type in preset {name}. Exiting...")

        if determinize not in determinization_map:
            raise ValueError(f"{determinize} is not a valid determinization strategy in preset {name}. Exiting...")

        if impurity not in impurity_map:
            raise ValueError(f"{impurity} is not a valid impurity measure in preset {name}. Exiting...")

        # Handle necessary cases which make the user interface simple
        label_pre_processor = None
        early_stopping = False
        if determinize in ['minnorm', 'random']:
            label_pre_processor = determinization_map[determinize](None)
            determinize = "none"
        if (determinize in ['maxfreq', 'auto']):
            early_stopping = True

        # determinizer must be auto when using multilablestuff
        if 'multilabel' in impurity:
            if not determinize == "auto":
                logging.error(f"{impurity} impurity measure automatically determinizes. Please use 'determinize: auto' when defining the preset.")
                sys.exit(-1)

        # if auto is used with any other, then give info message saying we use maxfreq
        if 'multilabel' not in impurity:
            if determinize == "auto":
                logging.info(f"INFO: Using the recommended maxfreq determinizer since the preset contained 'determinize: auto'.")
                determinize = "maxfreq"

        # if using logreg/svm/oc1, then determinizer must be passed to the split
        splitting_strategy = []
        for sp in combined_split:
            if sp in ['linear-logreg', 'linear-linsvm', 'oc1']:
                splitting_strategy.append(splitting_map[sp](determinization_map[determinize](None)))
            else:
                splitting_strategy.append(splitting_map[sp](None))

        impurity_measure = impurity_map[impurity](determinization_map[determinize](None))

        classifier = DecisionTree(splitting_strategy, impurity_measure, name,
                                  early_stopping=early_stopping, label_pre_processor=label_pre_processor)

        if safe_pruning:
            logging.info(f"Enabling safe pruning for preset {name}")
            classifier = SafePruning(classifier)

        # returns a flattened list
        return classifier

    def clean_parser(args):
        if args.output_cache:
            clear_output_cache()
        if args.run_cache:
            clear_run_cache()
        if args.all:
            clear_output_cache()
            clear_run_cache()
        sys.exit()

    def clear_run_cache():
        logging.info("Clearing default benchmark files 'benchmark.html' and 'benchmark.json'...")
        if not exists('benchmark.html'):
            logging.warning("The file 'benchmark.html' does not exist. Please run dtcontrol "
                            "from the working directory or clear the run cache manually.")
        if not exists('benchmark.json'):
            logging.warning("The file 'benchmark.json' does not exist. Please run dtcontrol "
                            "from the working directory or clear the run cache manually.")
            sys.exit()
        try:
            remove('benchmark.html')
            remove('benchmark.json')
        except:
            logging.error("Error deleting 'benchmark.html' and 'benchmark.json'. Please delete manually.")
            sys.exit(-1)

    def clear_output_cache():
        logging.info("Clearing default output cache folder 'decision_trees'...")
        if not exists('decision_trees'):
            logging.warning("The folder 'decision_trees' does not exist. Please run dtcontrol "
                            "from the working directory or clear the output manually.")
            sys.exit()
        try:
            shutil.rmtree('decision_trees')
        except:
            logging.error("Error deleting 'decision_trees'. Please delete manually.")
            sys.exit(-1)

    def core_parser(args):
        kwargs = dict()

        if args.input:
            dataset = args.input
        else:
            parser.print_help()
            sys.exit()

        kwargs["timeout"] = 2 * 60 * 60
        if args.timeout:
            kwargs["timeout"] = parse_timeout(args.timeout)

        if args.benchmark_file:
            filename, file_extension = splitext(args.benchmark_file)
            kwargs["benchmark_file"] = filename
        else:
            kwargs["benchmark_file"] = 'benchmark'
            logging.info("--benchmark-file/-b was not set. Defaulting to use 'benchmark.json'")

        if args.output:
            try:
                makedirs(args.output, exist_ok=True)
                kwargs["output_folder"] = args.output
            except PermissionError:
                sys.exit("Ensure permission exists to create output directory")

        kwargs["rerun"] = args.rerun
        if not args.rerun and isfile(kwargs["benchmark_file"]):
            logging.warning(
                f"Dataset - method combinations whose results are already present in '{kwargs['benchmark_file']}' "
                f"would not be re-run. Use the --rerun flag if this is what is desired.")

        suite = BenchmarkSuite(**kwargs)
        suite.add_datasets(dataset)

        # Parse config files
        default_config: OrderedDict = load_default_config()
        user_config: Union[None, OrderedDict] = None

        if args.config:
            try:
                yaml = YAML()
                user_config = yaml.load(open(args.config))
            except FileNotFoundError:
                sys.exit(
                    f"Error finding the config file. Please check if the file '{args.config}' exists in the current directory")
            except ScannerError:
                sys.exit(f"Scan error in the YAML configuration file '{args.config}'. Please re-check syntax.")

        classifiers = []
        run_config_table = []
        Row = namedtuple('Row',
                         ['Name', 'NumericPredicate', 'CategoricalPredicate', 'Determinize', 'Impurity', 'Tolerance',
                          'SafePruning'])

        if args.use_preset:
            if "all-user" in args.use_preset:
                presets = fetch_presets_from_loaded_config(user_config)
                if not presets:
                    sys.exit("Please specify a user config file via the --config switch. See help for more details.")
            elif "all-system" in args.use_preset:
                presets = fetch_presets_from_loaded_config(default_config)
            elif "all" in args.use_preset:
                presets = fetch_presets_from_loaded_config(user_config) \
                          + fetch_presets_from_loaded_config(default_config)
            else:
                presets = args.use_preset
            for preset in presets:
                numeric_split, categorical_split, determinize, impurity, tolerance, safe_pruning = get_preset(preset,
                                                                                                              user_config,
                                                                                                              default_config)
                classifiers.append(
                    get_classifier(numeric_split, categorical_split, determinize, impurity, tolerance=tolerance,
                                   safe_pruning=safe_pruning, name=preset))
                run_config_table.append(
                    Row(Name=preset, NumericPredicate=numeric_split, CategoricalPredicate=categorical_split,
                        Determinize=determinize, Impurity=impurity, Tolerance=tolerance,
                        SafePruning=safe_pruning))

        if not classifiers:
            logging.info("INFO: Assuming --use-preset default, since no preset explicitly specified.")
            preset = "default"
            numeric_split, categorical_split, determinize, impurity, tolerance, safe_pruning = get_preset(preset,
                                                                                                          user_config,
                                                                                                          default_config)
            classifiers.append(
                get_classifier(numeric_split, categorical_split, determinize, impurity, tolerance=tolerance,
                               safe_pruning=safe_pruning, name=preset))
            run_config_table.append(
                Row(Name=preset, NumericPredicate=numeric_split, CategoricalPredicate=categorical_split,
                    Determinize=determinize, Impurity=impurity, Tolerance=tolerance,
                    SafePruning=safe_pruning))


        logging.info("The following configurations would now be run:\n")
        logging.info(tabulate(run_config_table,
                              ['name', 'numeric-predicates', 'categorical-predicates', 'determinize', 'impurity',
                               'tolerance', 'safe-pruning'],
                              tablefmt="presto"))
        logging.info("\n")

        suite.benchmark(classifiers)

    logging.basicConfig(level=logging.INFO, format='%(message)s')

    parser = argparse.ArgumentParser(prog="dtcontrol",
                                     formatter_class=argparse.RawDescriptionHelpFormatter,
                                     description="Scroll to the end of the help message for Quick Start.",
                                     epilog="Examples:\n"
                                            "Create a file storing run configurations\n"
                                            "    dtcontrol preset --sample > user-config.yml\n\n"
                                            "Display all presets available with dtcontrol\n"
                                            "    dtcontrol preset --config user-config.yml --list\n\n"
                                            "Run the 'my-config' preset on the SCOTS model located at 'examples/cartpole.scs'\n\n"
                                            "    dtcontrol --input examples/cartpole.scs --config user-config.yml --use-preset my-config")
    parser.set_defaults(func=core_parser)

    version = pkg_resources.require("dtcontrol")[0].version
    parser.add_argument("-v", "--version", action='version',
                        version=f'%(prog)s {version}')

    input_output = parser.add_argument_group('input/output')
    input_output.add_argument("--input", "-i", nargs="+", type=(lambda x: is_valid_file_or_folder(parser, x)),
                              help="The input switch takes in one or more space separated file names or "
                                   "a folder name which contains valid controllers (.scs, .dump or .csv)")

    input_output.add_argument("--output", "-o", type=str,
                              help="The output switch takes in a path to a folder where the constructed controller "
                                   "representation would be saved (c and dot)")

    input_output.add_argument("--benchmark-file", "-b", metavar="FILENAME", type=str,
                              help="Saves statistics pertaining the construction of the decision trees and their "
                                   "sizes into a JSON file, and additionally allows to view it via an HTML file.")

    run_config = parser.add_argument_group('run configurations')
    run_config.add_argument("--config", "-c", metavar="CONFIGFILE", type=str,
                            help="Specify location of a YAML file containing run configurarions. Use along with the "
                                 "--use-preset switch. More details in the User Manual.")

    run_config.add_argument("--use-preset", "-p", type=str, nargs="+",
                            help="Run one or more presets defined in the CONFIGFILE. If the --config switch has not "
                                 "been used, then presets are chosen from the system-level configuration file. Special "
                                 "parameters for this switch include 'all', 'all-user', 'all-system'. Refer the User "
                                 "Manual for more details.")

    run_config.add_argument("--rerun", "-r", action='store_true',
                            help="Rerun the experiment for all input-method combinations. Overrides the default "
                                 "behaviour of not running benchmarks for combinations which are already present"
                                 " in the benchmark file.")

    run_config.add_argument("--timeout", "-t", type=str,
                            help="Sets a timeout for each method. Can be specified in seconds, minutes "
                                 "or hours (eg. 300s, 7m or 3h)")

    subparsers = parser.add_subparsers(title='other commands',
                                       # description = 'Supplementary commands',
                                       help='Run \'dtcontrol COMMAND --help\' to see command specific help')

    parser_conf = subparsers.add_parser(name="preset")
    parser_conf.add_argument("--config", "-c", metavar="CONFIGFILE", type=str,
                             help="Specify location of a YAML file containing run configurarions")
    parser_conf.add_argument("--list", "-l", action='store_true',
                             help="List all available presets (run configurations)")
    parser_conf.add_argument("--sample", "-s", action='store_true',
                             help="Print a sample user configuration file")
    parser_conf.set_defaults(func=preset_parser)

    parser_clean = subparsers.add_parser(name="clean")
    parser_clean.add_argument("--all", "-a", action='store_true',
                              help="Clear cache and outputs")
    parser_clean.add_argument("--run-cache", "-c", action='store_true',
                              help="Clear run cache")
    parser_clean.add_argument("--output-cache", "-o", action='store_true',
                              help="Clear output folder")
    parser_clean.set_defaults(func=clean_parser)

    args = parser.parse_args()
    args.func(args)


if __name__ == "__main__":
    main()



import math
import pickle
import random

from dataset.multi_output_dataset import MultiOutputDataset
# from dd import cudd as _bdd
from dd import autoref as _bdd  # TODO: Might use cudd
from tqdm import tqdm

class BDD:
    """
    Standard BDD classifier, using the dd python bindings for cudd (because sylvan was broken when I tried, 17.10.19)
    """

    def __init__(self):
        self.name = 'BDD'
        self.bdd = _bdd.BDD()
        self.bdd.configure(reordering=False)
        self.bb_table = dict()
        self.name2node = dict()
        self.result = None

    @staticmethod
    def is_applicable(dataset):
        return True
        # TODO: Add support for multi sets

    def get_stats(self):
        return {"nodes": len(self.result)}

    def predict(self, dataset):
        # TODO
        # check that everything in X_train is predicted correctly
        # predict_state tell us the allowed actions for one state
        # Iterate over whole dataset and use it
        pass

    def predict_state(self, row, X_vars, Y_vars):
        # Get bit_repr of all states vars
        state_bit_values = dict()
        for i in range(0, len(X_vars)):
            var_name = X_vars[i]
            value = row[i]
            num_bits, precision_multiplier, signed = self.bb_table[var_name]
            if signed and value < 0:
                state_bit_values[var_name + str(num_bits - 1)] = 1
                value = value * -1
                num_bits -= 1  # do not consider the most significant bit anymore
            value = math.floor(value * precision_multiplier)  # use the decimal approach (described below)
            for i in range(0, num_bits):
                if value - math.pow(2, i) >= 0:
                    state_bit_values[var_name + str(i)] = True
                    value -= math.pow(2, i)
                else:
                    state_bit_values[var_name + str(i)] = False

        allowed_actions = set()
        # try all combinations for Y_vars, remember those that return true
        # TODO: This is too much work right now.

    def check_result(self, state_bit_values, node):
        # Base cases
        if node == self.bdd.false:
            return False
        if node == self.bdd.true:
            return True
        # State variable: Read value and continue accordingly
        if node.var in state_bit_values.keys():
            if state_bit_values[node.var]:
                return self.check_result(state_bit_values, node.low if node.negated else node.high)
            else:
                return self.check_result(state_bit_values, node.high if node.negated else node.low)
        # # Action variable: Return all possible combinations
        # if node.var not in state_bit_values.keys():
        #     if_one = self.check_result(state_bit_values, node.low if node.negated else node.high)
        #     if_zero = self.check_result(state_bit_values, node.high if node.negated else node.low)
        #     if if_one is None and if_zero is None:
        #         return None
        #
        #     if if_one == []:
        #         result = [f"{node.var}=1"]
        #     else:
        #         result = list(itertools.product([f"{node.var}=1"], if_one))
        #     if if_zero == []:
        #         result += [f"{node.var}=0"]
        #     else:
        #         result += list(itertools.product([f"{node.var}=0"], if_one))
        #     return result

    def save(self, filename):
        # TODO: Does this work?
        with open(filename, 'wb') as outfile:
            pickle.dump(self, outfile)

    # TODO: Determinize first and then BDD; need to compare to this as well
    # TODO: check whether there is some cool cudd stuff that we can utilize
    def fit(self, dataset):
        # fill bb_table, the almighty table that contains the semantics and number of bits for all vars
        self.add_to_bb_table(dataset.X_metadata)
        self.add_to_bb_table(dataset.Y_metadata)
        # initialize bdd and name2node based on bb_table
        for var in self.bb_table.keys():
            for i in range(0, self.bb_table[var][0]):
                blasted_name = var + str(i)
                self.bdd.declare(blasted_name)
                self.name2node[blasted_name] = self.bdd.var(blasted_name)
        self.reorder_randomly()

        # For each state: construct the sub_result_state for the x's
        # Then construct subresult_sa_pair as OR_{action in allowed_actions} sub_result_state AND action
        # Then or this with the whole result
        row_num = -1
        total = len(dataset.X_train)
        for row in tqdm(dataset.X_train):
            row_num += 1
            # This assumes that the variables in X_train are in the same order as X_vars
            sub_result_state = self.make_sub_result_state(row, dataset.X_metadata["variables"])

            sub_result_nondet_initialized = False  # need this to initalize sub_result_sa
            multi = isinstance(dataset, MultiOutputDataset)
            # for every nondeterministic choice
            for i in range(0, len(dataset.Y_train[row_num]) if not multi else len(dataset.Y_train[0][row_num])):
                for j in range(0, len(dataset.Y_metadata["variables"])):
                    action_to_add = dataset.Y_train[row_num][i] if not multi else dataset.Y_train[j][row_num][i]
                    if action_to_add == -1:
                        continue  # -1 does not represent a playable action

                    # and all the actions (or the single action if not multi output) to sub_result_state
                    if j == 0:
                        sub_result_action = self.bdd.apply('and', sub_result_state,
                                                           self.bdd_for(dataset.index_to_value[action_to_add],
                                                                        dataset.Y_metadata["variables"][j]))
                    else:
                        sub_result_action = self.bdd.apply('and', sub_result_action,
                                                           self.bdd_for(dataset.index_to_value[action_to_add],
                                                                        dataset.Y_metadata["variables"][j]))

                # Now sub_result_action represents the and of all actions for a single nondet choice and the state
                if not sub_result_nondet_initialized:
                    sub_result_nondet = sub_result_action
                    sub_result_nondet_initialized = True
                else:
                    sub_result_nondet = self.bdd.apply('or', sub_result_nondet, sub_result_action)

            if row_num == 0:
                self.result = sub_result_nondet
            else:
                self.result = self.bdd.apply('or', self.result, sub_result_nondet)

        # collect garbage and reorder heuristics
        print("Before collecting garbage: result %s, BDD %s" % (len(self.result), len(self.bdd)))
        self.bdd.collect_garbage()
        print("After: result %s, BDD %s" % (len(self.result), len(self.bdd)))

        # reorder with dd until convergence
        i = 0
        while True:
            print(str(i) + ": result %s, BDD %s" % (len(self.result), len(self.bdd)))
            i += 1
            bdd_size = len(self.bdd)
            _bdd.reorder(self.bdd)
            if bdd_size == len(self.bdd):
                print("Reordering did not change size, local optimum BDD computed.")
                break

    """
    Returns a dict of the form
    varname -> num_bits, precision_multiplier, signed
    num_bits is how many bits we use to represent the var
    we use stepsize of the var (from metadata) to decide how precise to save a number
    Then we compute precision_multiplier based on that.
    Example: step size 0.08; then precision multiplier is 100, and we cut off anything beyond two decimals
    In general, precision_multiplier is 10^{-floor(log_10 (step size))}
    if signed, we always use the most significant bit (highest index, namely num_bits-1) to represent the sign
        (not 2 complement or anything fancy for negative numbers, 0 exists twice but should always be positive)
        
    semantics is:
    number to this special binary format:
    If signed, most significant bit gives sign and is ignored for number; else all bits are for number
    multiply number by precision_multiplier and cut everything after the comma (floor)
    represent this multiplied integer as bits.
    special binary format to number:
    If signed, most significant bit gives sign and is ignored for number, else all bits are for number
    get integer represented by bits
    divide by precision_multiplier to get actual number (up to necessary precision)
    """

    def add_to_bb_table(self, metadata):
        for i in range(0, len(metadata["variables"])):
            name = metadata["variables"][i]
            signed = True if metadata["min"][i] < 0 else False
            # this is what we multiply with to be precise up to step size; if step size > 1, we make it 1
            if metadata["step_size"][i] >= 1:
                precision_multiplier = 1
            else:
                precision_multiplier = pow(10, - math.floor(math.log(metadata["step_size"][i], 10)))
            largest_number = max(math.fabs(metadata["min"][i]), math.fabs(metadata["max"][i]))
            if largest_number == 0 or precision_multiplier == 0:  # catch some weird corner cases
                num_bits = 1 + (1 if signed else 0)
            else:  # Need the 1+ in the beginning, off-by-1 error. Try it for 8 and 10, you will see.
                num_bits = 1 + int(math.log(largest_number * precision_multiplier, 2) + (1 if signed else 0))
            self.bb_table[metadata["variables"][i]] = (num_bits, precision_multiplier, signed)

    # This assumes that the variables in X_train are in the same order as X_vars
    def make_sub_result_state(self, row, X_vars):
        for i in range(0, len(X_vars)):
            sub_result = self.bdd_for(row[i], X_vars[i]) if i == 0 else \
                self.bdd.apply('and', sub_result, self.bdd_for(row[i], X_vars[i]))
        return sub_result

    def bdd_for(self, value, varname):
        num_bits, precision_multiplier, signed = self.bb_table[varname]
        remainder = math.floor(value * precision_multiplier)  # transform to our special binary format

        # go over value in reverse order, set bits in BDD
        for i in reversed(range(0, num_bits)):
            bit_varname = varname + str(i)
            # handle corner case signed: then most significant bit (num_bits-1) is treated separately
            if signed and i == num_bits - 1:
                sub_result = self.name2node[bit_varname] if value < 0 else self.bdd.apply('not',
                                                                                          self.name2node[bit_varname])
                remainder = remainder * -1 if remainder < 0 else remainder
                continue
            # check whether the i-th bit is set by subtracting 2^{its exponent}

            if remainder - pow(2, i) >= 0:
                remainder -= pow(2, i)
                # if first bit, initialize subres; else 'and' subres and new thing
                sub_result = self.name2node[bit_varname] if i == num_bits - 1 \
                    else self.bdd.apply('and', sub_result, self.name2node[bit_varname])
            else:  # Same as above, but with a not in front of the node
                sub_result = self.bdd.apply('not', self.name2node[bit_varname]) if i == num_bits - 1 \
                    else self.bdd.apply('and', sub_result, self.bdd.apply('not', self.name2node[bit_varname]))

        if remainder != 0:
            raise Exception(f"Fatal error: Bitblasting variable {varname} did not succeed with value {value}.")

        return sub_result

    # This seems such a bad way, but all others break somehow...
    def reorder_randomly(self):
        my_order = dict()
        all_names = []
        for name in self.name2node.keys():
            all_names += [name]

        for i in range(0, len(all_names)):
            c = random.choice(all_names)
            my_order[c] = i
            all_names.remove(c)

        _bdd.reorder(self.bdd, my_order)
        # print("OrderStart:")
        # for item in sorted(self.bdd.vars):
        #    print("%s: %s" % (item, self.bdd.vars[item]))

    def export_dot(self, file=None):
        pass

    def export_c(self, file=None):
        # TODO
        # Use the ITE code from old notebook; using functions for all is not nice code, but that's why it's auto generated.
        # Anyway we should compare only the compiled versions.
        pass

    def export_vhdl(self, file=None):
        pass


#ds = SingleOutputDataset("../dumps/cruise-small-latest.dump")
#ds = MultiOutputDataset("../XYdatasets/10rooms.scs")
# ds = SingleOutputDataset("../XYdatasets/cartpole.scs")
# ds.load_if_necessary()
# bdd_classifier = BDD()
# bdd_classifier.fit(ds)



import random

import numpy as np

from dtcontrol.decision_tree.pre_processing.label_pre_processor import LabelPreProcessor

class RandomPreProcessor(LabelPreProcessor):
    """
    Pre-processes the labels randomly.
    """

    def __init__(self):
        super().__init__()

    def preprocess_single_output(self, dataset):
        return np.apply_along_axis(lambda x: np.random.choice(x[x != -1]),
                                   axis=1,
                                   arr=dataset.y).reshape((-1, 1))

    def preprocess_multi_output(self, dataset):
        zipped = np.stack(dataset.y, axis=2)
        result = []
        for arr in zipped:
            choices = [t for t in arr if t[0] != -1]
            result.append(random.choice(choices))
        return np.array(np.split(np.array(result), len(dataset.y), 1))



from abc import ABC, abstractmethod

class BenchmarkSuiteClassifier(ABC):
    def __init__(self, name):
        self.name = name

    def get_name(self):
        return self.name

    @abstractmethod
    def is_applicable(self, dataset):
        pass

    @abstractmethod
    def fit(self, dataset):
        """
        Trains the classifier on the given dataset.
        """
        pass

    @abstractmethod
    def predict(self, dataset, actual_values=True):
        """
        Classifies a dataset.
        :param dataset: the dataset to classify
        :param actual_values: if True, the actual float values are predicted. if False, the index labels are predicted.
        :return: a list of the predicted labels.
        """
        pass

    @abstractmethod
    def get_stats(self):
        """
        Returns a dictionary of statistics to be saved and displayed (e.g. the number of nodes in the tree).
        :return: the dictionary of statistics
        """
        pass

    @abstractmethod
    def save(self, file):
        """
        Saves the classifier to a file (for debugging purposes).
        """
        pass

    @abstractmethod
    def print_dot(self, x_metadata, y_metadata):
        """
        Prints the classifier in the dot (graphviz) format.
        :param x_metadata: metadata describing the columns of the data
        :param y_metadata: metadata describing the columns of the labels
        :return: the dot string
        """
        pass

    @abstractmethod
    def print_c(self):
        """
        Prints the classifier as nested if-else statements in the C syntax.
        :return: the C string
        """
        pass



import unittest

from dtcontrol.dataset.multi_output_dataset import MultiOutputDataset
from dtcontrol.decision_tree.decision_tree import DecisionTree
from dtcontrol.decision_tree.impurity.entropy import Entropy
from dtcontrol.decision_tree.splitting.axis_aligned import AxisAlignedSplittingStrategy
from dtcontrol.decision_tree.splitting.categorical_multi import CategoricalMultiSplittingStrategy

class TestDot(unittest.TestCase):
    def test_dot(self):
        try:
            ds = MultiOutputDataset('golf_multi.csv')
            ds.load_if_necessary()
        except FileNotFoundError:
            ds = MultiOutputDataset('test/golf_multi.csv')
            ds.load_if_necessary()

        dt = DecisionTree([CategoricalMultiSplittingStrategy(), AxisAlignedSplittingStrategy()],
                          Entropy(), 'categorical')
        dt.fit(ds)

        try:
            with open('golf_multi.dot') as infile:
                expected = infile.read()
        except FileNotFoundError:
            with open('test/golf_multi.dot') as infile:
                expected = infile.read()
        self.assertEqual(expected, dt.print_dot(ds.x_metadata, ds.y_metadata))



import logging
import os
import re
import shutil
import subprocess
import sys
from os.path import exists

import numpy as np

import dtcontrol
import dtcontrol.globals
from dtcontrol.decision_tree.determinization.non_determinizer import NonDeterminizer
from dtcontrol.decision_tree.splitting.axis_aligned import AxisAlignedSplit
from dtcontrol.decision_tree.splitting.linear_split import LinearSplit
from dtcontrol.decision_tree.splitting.splitting_strategy import SplittingStrategy
from dtcontrol.util import log_without_newline

class OC1SplittingStrategy(SplittingStrategy):
    def __init__(self, determinizer=NonDeterminizer(), num_restarts=10, num_jumps=5, delete_tmp=True):
        self.determinizer = determinizer
        self.oc1_path = 'decision_tree/OC1_source/mktree'
        self.header_file = 'decision_tree/OC1_source/oc1.h'
        self.tmp_path = '.dtcontrol_tmp'
        self.output_file = f'{self.tmp_path}/output'
        self.data_file = f'{self.tmp_path}/data.csv'
        self.dt_file = f'{self.tmp_path}/dt'
        self.log_file = f'{self.tmp_path}/log'
        self.num_restarts = num_restarts
        self.num_jumps = num_jumps
        self.delete_tmp = delete_tmp
        self.regex_pattern = re.compile(r'IMPURITY (.+) \(\)')
        if not os.path.exists(self.oc1_path):
            self.compile_oc1()

    def compile_oc1(self):
        for path in dtcontrol.__path__:
            oc1_src = f"{path}/decision_tree/OC1_source"
            if os.path.exists(oc1_src):
                if os.path.exists(oc1_src + "/mktree"):
                    self.oc1_path = oc1_src + "/mktree"
                    return
                try:
                    log_without_newline("Compiling OC1... ")
                    subprocess.call("make", cwd=oc1_src)
                    self.oc1_path = oc1_src + "/mktree"
                    logging.info("Compiled OC1")
                except subprocess.CalledProcessError:
                    logging.error("Compiling OC1 failed")
                    sys.exit(-1)
            else:
                logging.error("Could not find OC1 files")
                sys.exit(-1)

    def find_split(self, dataset, impurity_measure):
        x_numeric = dataset.get_numeric_x()
        if x_numeric.shape[1] == 0:
            return None
        y = self.determinizer.determinize(dataset)
        if not os.path.exists(self.tmp_path):
            os.mkdir(self.tmp_path)
        self.save_data_to_file(x_numeric, y)
        self.set_impurity_measure(impurity_measure)
        self.execute_oc1()
        split = self.parse_oc1_dt(dataset)
        if self.delete_tmp:
            shutil.rmtree(self.tmp_path)
        return split

    def set_impurity_measure(self, impurity_measure):
        for path in dtcontrol.__path__:
            header = f'{path}/decision_tree/OC1_source/oc1.h'
            if os.path.exists(header):
                self.header_file = header
        with open(self.header_file, 'r') as infile:
            content = infile.read()
        match = re.search(self.regex_pattern, content)
        assert match
        if match.group(1) == impurity_measure.get_oc1_name():
            return
        new_content = re.sub(self.regex_pattern, f'IMPURITY {impurity_measure.get_oc1_name()} ()', content)
        with open(self.header_file, 'w') as outfile:
            outfile.write(new_content)
        logging.disable(logging.INFO)
        self.compile_oc1()
        logging.disable(logging.NOTSET)

    def save_data_to_file(self, x, y):
        data = np.c_[x, y]
        num_float_columns = data.shape[1] - 1
        np.savetxt(self.data_file, data, fmt=' '.join(['%f'] * num_float_columns + ['%d']), delimiter='\t')

    def execute_oc1(self):
        command = f'{self.oc1_path} -t {self.data_file} -D {self.dt_file} -p0 -i{self.num_restarts} ' \
                  f'-j{self.num_jumps} -l {self.log_file}'
        with open(self.output_file, 'w+') as out:
            p = subprocess.Popen(command.split(' '), stdout=out)
            dtcontrol.globals.oc1_pid = p.pid
            p.wait()
            dtcontrol.globals.oc1_pid = None

    def parse_oc1_dt(self, dataset):
        if not exists(self.dt_file):
            return None
        with open(self.dt_file) as infile:
            while not infile.readline().startswith('Root'):
                pass
            hyperplane_str = infile.readline()

        summands = hyperplane_str[:-4].split(' + ')
        intercept = float(summands[-1])
        dim = dataset.get_numeric_x().shape[1]
        j = 0
        coefficients = []
        for i in range(1, dim + 1):
            if f"x[{i}]" in summands[j]:
                coefficients.append(float(summands[j].split()[0]))
                j = j + 1
            else:
                coefficients.append(0)
        if len([c for c in coefficients if c != 0]) == 1:
            for i in range(len(coefficients)):
                if coefficients[i] != 0:
                    return AxisAlignedSplit(dataset.map_numeric_feature_back(i), -intercept)
        real_coefficients = LinearSplit.map_numeric_coefficients_back(coefficients, dataset)
        return LinearSplit(coefficients, intercept, real_coefficients, dataset.numeric_columns)



import logging

import numpy as np
import pandas as pd

from dtcontrol.dataset.dataset_loader import DatasetLoader

class CSVDatasetLoader(DatasetLoader):
    def _load_dataset(self, filename):
        with open(filename, 'r') as f:
            logging.info(f"Reading from {filename}")
            f.readline()  # whether permissive

            state_dim, input_dim = map(int, f.readline().split("BEGIN")[1].split())

            ds = pd.read_csv(f, header=None)

            unique_list = []
            for i in range(state_dim, state_dim + input_dim):
                unique_list += ds[i].unique().tolist()
            index_to_actual = {x + 1: y for x, y in enumerate(set(unique_list))}
            value_to_index = {y: x for x, y in index_to_actual.items()}

            ds[[i for i in range(state_dim, state_dim + input_dim)]] = ds[
                [i for i in range(state_dim, state_dim + input_dim)]].applymap(lambda x: value_to_index[x])

            grouped = ds.groupby([i for i in range(state_dim)], sort=False)
            aggregate = grouped[state_dim].apply(list).reset_index(name=state_dim)
            for i in range(1, input_dim):
                aggregate[state_dim + i] = grouped[state_dim + i].apply(list).reset_index(name=state_dim + i)[
                    state_dim + i]

            max_non_det = aggregate[state_dim].agg(len).max()

            for i in range(0, input_dim):
                aggregate[state_dim + i] = aggregate[state_dim + i].apply(
                    lambda ls: ls + [-1 for i in range(max_non_det - len(ls))])

            x = np.array(aggregate[[i for i in range(state_dim)]])

            if input_dim > 1:
                y = np.ndarray((input_dim, x.shape[0], max_non_det), dtype=np.int16)
                for i in range(input_dim):
                    y[i] = np.array(aggregate[state_dim + i].tolist())
            else:  # input_dim = 1
                y = np.array(aggregate[state_dim].tolist())

            # construct metadata
            # assumption is that UPPAAL only works with integers
            x_metadata = dict()
            x_metadata["variables"] = [f"x_{i}" for i in range(state_dim)]
            x_metadata["categorical"] = []
            x_metadata["min"] = [int(i) for i in np.amin(x, axis=0)]
            x_metadata["max"] = [int(i) for i in np.amax(x, axis=0)]
            x_metadata["step_size"] = None  # todo

            y_metadata = dict()
            y_metadata["variables"] = [f"u_{i}" for i in range(input_dim)]
            y_metadata["min"] = [min(index_to_actual.values())]
            y_metadata["max"] = [max(index_to_actual.values())]
            y_metadata["step_size"] = None  # todo

            logging.debug(x_metadata)
            logging.debug(y_metadata)

            return (x, x_metadata, y, y_metadata, index_to_actual)



from unittest import TestCase

from dtcontrol.decision_tree.decision_tree import Node, DecisionTree
from dtcontrol.decision_tree.impurity.entropy import Entropy
from dtcontrol.post_processing.safe_pruning import SafePruning

class TestSafePruning(TestCase):
    def test_run(self):
        lll = self.create_leaf([1, 2, 3])
        llr = self.create_leaf([2, 3])
        ll = self.create_parent(lll, llr)
        lr = self.create_leaf(2)
        l = self.create_parent(ll, lr)
        rl = self.create_leaf([1, 2, 3])
        rr = self.create_leaf([4, 5])
        r = self.create_parent(rl, rr)
        root = self.create_parent(l, r)

        sp = SafePruning(DecisionTree([], Entropy(), 'name'))
        sp.classifier.root = root
        sp.run()

        self.assertEqual(5, root.num_nodes)
        self.assertTrue(root.children[0].is_leaf())
        self.assertEqual(2, root.children[0].index_label)
        self.assertTrue(root.children[1].children[0].is_leaf())
        self.assertEqual([1, 2, 3], root.children[1].children[0].index_label)
        self.assertTrue(root.children[1].children[1].is_leaf())
        self.assertEqual([4, 5], root.children[1].children[1].index_label)

    def test_rounds(self):
        ll = self.create_leaf([1,2])
        lr = self.create_leaf([3,4])
        l = self.create_parent(ll, lr)
        rll = self.create_leaf([1,2,3])
        rlr = self.create_leaf([2,3,4])
        rl = self.create_parent(rll, rlr)
        rrl = self.create_leaf([2,5])
        rrr = self.create_leaf([8, 2])
        rr = self.create_parent(rrl, rrr)
        r = self.create_parent(rl, rr)
        root = self.create_parent(l, r)

        sp = SafePruning(DecisionTree([], Entropy(), 'name'), rounds=1)
        sp.classifier.root = root
        sp.run()

        self.assertEqual(7, root.num_nodes)
        self.assertTrue(root.children[0].children[0].is_leaf())
        self.assertEqual([1,2], root.children[0].children[0].index_label)
        self.assertTrue(root.children[0].children[1].is_leaf())
        self.assertEqual([3, 4], root.children[0].children[1].index_label)
        self.assertTrue(root.children[1].children[0].is_leaf())
        self.assertEqual([2,3], root.children[1].children[0].index_label)
        self.assertTrue(root.children[1].children[1].is_leaf())
        self.assertEqual(2, root.children[1].children[1].index_label)

    @staticmethod
    def create_leaf(label):
        node = Node(None, None)
        node.num_nodes = 1
        node.index_label = label
        node.actual_label = label
        return node

    @staticmethod
    def create_parent(left, right):
        node = Node(None, None)
        node.children = [left, right]
        node.num_nodes = 1 + left.num_nodes + right.num_nodes
        return node



import logging
from os.path import splitext, exists

import numpy as np

from dtcontrol.dataset.dataset_loader import DatasetLoader

class PrismDatasetLoader(DatasetLoader):
    def _load_dataset(self, filename):
        logging.info(f"Reading from {filename}")
        name, ext = splitext(filename)
        states_file = name + '_states.prism'
        if not exists(states_file):
            raise ValueError(f'Couldn\'t find states file for prism dataset {filename}.')

        state_mapping = {}
        with open(states_file, 'r') as f:
            x_variables = f.readline()
            x_variables = x_variables.replace('(', '').replace(')', '').split(',')
            x_variables = [x.strip() for x in x_variables]
            last_state = -1
            for line in f:
                n, t = line.split(':')
                last_state += 1
                assert int(n) == last_state

                def to_int(x):
                    if x == 'true' or x == 'false':
                        return int(x == 'true')
                    return int(x)

                state_mapping[int(n)] = [to_int(i) for i in t.replace('(', '').replace(')', '').split(',')]

        x = []
        y = []
        y_variables = {}
        y_index = 1  # labels start with 1
        with open(filename, 'r') as f:
            for line in f:
                n, a = line.split(':')
                x.append(state_mapping[int(n)])
                if a not in y_variables:
                    y_variables[a] = y_index
                    y_index += 1
                y.append([y_variables[a]])

        x_metadata = dict()
        x_metadata["variables"] = x_variables
        x_metadata["categorical"] = []
        x_metadata["min"] = [int(i) for i in np.amin(x, axis=0)]
        x_metadata["max"] = [int(i) for i in np.amax(x, axis=0)]
        x_metadata["step_size"] = None  # todo

        index_to_actual = {i: i - 1 for i in y_variables.values()}

        y_metadata = dict()
        y_metadata["categorical"] = [0]  # we always have exactly one output, and it is categorical
        y_metadata["category_names"] = {0: [v.strip() for v in sorted(y_variables.keys(), key=y_variables.get)]}
        y_metadata["min"] = [min(index_to_actual.values())]
        y_metadata["max"] = [max(index_to_actual.values())]
        y_metadata["step_size"] = None  # todo

        logging.debug(x_metadata)
        logging.debug(y_metadata)

        return (np.array(x), x_metadata, np.array(y), y_metadata, index_to_actual)



import sys

import numpy as np

from dtcontrol.decision_tree.impurity.deterministic_impurity_measure import DeterministicImpurityMeasure

class SumMinority(DeterministicImpurityMeasure):
    def calculate_impurity(self, dataset, split):
        if len(split.get_masks(dataset)) == 1:
            return sys.maxsize
        minorities = []
        for mask in split.get_masks(dataset):
            subset_labels = self.determinizer.determinize(dataset.from_mask_optimized(mask))
            if len(subset_labels) == 0:
                return sys.maxsize
            minorities.append(self.calculate_minority(subset_labels))
        return sum(minorities)

    @staticmethod
    def calculate_minority(y):
        label = np.bincount(y).argmax()
        return len(y[y != label])

    def get_oc1_name(self):
        return 'summinority'



import sys

import numpy as np

from dtcontrol.decision_tree.impurity.deterministic_impurity_measure import DeterministicImpurityMeasure

class EntropyRatio(DeterministicImpurityMeasure):
    def calculate_impurity(self, dataset, split):
        if any(np.all(mask == False) for mask in split.get_masks(dataset)) or \
                len(split.get_masks(dataset)) == 1:
            return sys.maxsize

        split_entropy = self.calculate_split_entropy(dataset, split)
        split_info = self.calculate_split_info(dataset, split)
        return split_entropy / split_info

    def calculate_split_entropy(self, dataset, split):
        entropy = 0
        for mask in split.get_masks(dataset):
            subset_labels = self.determinizer.determinize(dataset.from_mask_optimized(mask))
            entropy += (len(subset_labels) / len(dataset)) * EntropyRatio.calculate_entropy(subset_labels)
        assert entropy >= 0
        return entropy

    @staticmethod
    def calculate_entropy(y):
        num_labels = len(y)
        unique = np.unique(y)
        probabilities = [len(y[y == label]) / num_labels for label in unique]
        return sum(-prob * np.log2(prob) for prob in probabilities)

    def calculate_split_info(self, dataset, split):
        info = 0
        for mask in split.get_masks(dataset):
            subset_labels = self.determinizer.determinize(dataset.from_mask_optimized(mask))
            info -= (len(subset_labels) / len(dataset)) * np.log2((len(subset_labels) / len(dataset)))
        assert info > 0
        return info

    def get_oc1_name(self):
        return None



from dtcontrol.decision_tree.splitting.split import Split
from dtcontrol.decision_tree.splitting.splitting_strategy import SplittingStrategy

class CategoricalSingleSplittingStrategy(SplittingStrategy):
    def find_split(self, dataset, impurity_measure):
        x_categorical = dataset.get_categorical_x()
        splits = {}
        for feature in range(x_categorical.shape[1]):
            real_feature = dataset.map_categorical_feature_back(feature)
            for value in set(x_categorical[:, feature]):
                split = CategoricalSingleSplit(real_feature, value)
                splits[split] = impurity_measure.calculate_impurity(dataset, split)

        if not splits:
            return None
        return min(splits.keys(), key=splits.get)

class CategoricalSingleSplit(Split):
    """
    A split of the form feature == value.
    """

    def __init__(self, feature, value):
        self.feature = feature
        self.value = value

    def predict(self, features):
        v = features[:, self.feature][0]
        return 0 if v == self.value else 1

    def get_masks(self, dataset):
        mask = dataset.x[:, self.feature] == self.value
        return [mask, ~mask]

    def print_dot(self, variables=None, category_names=None):
        if variables:
            var = variables[self.feature]
        else:
            var = f'x[{self.feature}]'
        if category_names and self.feature in category_names:
            val = category_names[self.feature][self.value]
        else:
            val = self.value
        return f'{var} == {val}'

    def print_c(self):
        return f'x[{self.feature}] == {self.value}'

    def print_vhdl(self):
        return f'x{self.feature}] == {self.value}'



import numpy as np

from dtcontrol.decision_tree.pre_processing.label_pre_processor import LabelPreProcessor

class NormPreProcessor(LabelPreProcessor):
    """
    Pre-processes the labels with the (min/max)-norm approach.
    """

    def __init__(self, comp):
        """
        :param comp: the comparison function to be used, either min or max
        """
        super().__init__()
        self.comp = comp

    def preprocess_single_output(self, dataset):
        return np.apply_along_axis(lambda x: self.comp(x[x != -1], key=lambda i: dataset.index_to_actual[i] ** 2),
                                   axis=1,
                                   arr=dataset.y).reshape((-1, 1))

    def preprocess_multi_output(self, dataset):
        zipped = np.stack(dataset.y, axis=2)
        result = []
        for arr in zipped:
            result.append(self.comp([t for t in arr if t[0] != -1],
                                    key=lambda t: sum(dataset.index_to_actual[j] ** 2 for j in t)))
        return np.array(np.split(np.array(result), len(dataset.y), 1))



from os.path import join, exists, abspath, getsize
from urllib.parse import quote

from jinja2 import Environment, FileSystemLoader

import dtcontrol

GRAPHVIZ_URL = 'https://dreampuf.github.io/GraphvizOnline/#'
file_loader = FileSystemLoader([path + "/ui" for path in dtcontrol.__path__])
env = Environment(loader=file_loader)

class TableController:
    def __init__(self, html_file, output_folder, is_artifact=False):
        """
        :param html_file: the html file to save the table to
        :param output_folder: the folder containing the dot and C files
        :param is_artifact: if enabled, always produces a table exactly corresponding to the one in the paper
        """
        self.html_file = html_file
        self.output_folder = output_folder
        self.is_artifact = is_artifact

    @staticmethod
    def load_resource(filename):
        full_paths = [path + "/ui/" + filename for path in dtcontrol.__path__ if exists(path + "/ui/" + filename)]
        with open(full_paths[0]) as infile:
            script = infile.read()
        return script

    def update_and_save(self, results, last_run_datasets, last_run_classifiers):
        template = env.get_template('table.html')
        script_js = self.load_resource('table.js')
        jquery_js = self.load_resource('jquery-3.4.1.min.js')
        bootstrap_css = self.load_resource('bootstrap.min.css')
        bootstrap_js = self.load_resource('bootstrap.min.js')
        bootstrap_toggle_css = self.load_resource('bootstrap4-toggle.min.css')
        bootstrap_toggle_js = self.load_resource('bootstrap4-toggle.min.js')
        style_css = self.load_resource('style.css')

        if self.is_artifact:
            table, row_metadata, column_names = self.get_table_data_artifact(results)
        else:
            table, row_metadata, column_names = self.get_table_data(results)
        with open(self.html_file, 'w+') as out:
            out.write(template.render(
                column_names=column_names,
                row_metadata=row_metadata,
                table=table,
                links_table=self.get_dot_and_c_links(row_metadata, column_names),
                last_run_datasets=last_run_datasets,
                last_run_classifiers=last_run_classifiers,
                script_js=script_js,
                jquery_js=jquery_js,
                bootstrap_css=bootstrap_css,
                bootstrap_js=bootstrap_js,
                bootstrap_toggle_css=bootstrap_toggle_css,
                bootstrap_toggle_js=bootstrap_toggle_js,
                style_css=style_css
            ))

    @staticmethod
    def get_table_data(results):
        row_names = sorted(list(results.keys()))
        column_names = set()
        for dataset in results:
            column_names.update(results[dataset]['classifiers'].keys())
        column_names = sorted(list(column_names))
        table = []
        for dataset in row_names:
            row = []
            for classifier in column_names:
                if classifier in results[dataset]['classifiers']:
                    cell = results[dataset]['classifiers'][classifier]
                else:
                    cell = 'not yet computed'
                row.append(cell)
            table.append(row)
        row_metadata = [{'name': r, 'domain_of_controller': results[r]['metadata']['Y_metadata']['num_rows'],
                         'state_action_pairs': results[r]['metadata']['Y_metadata']['num_flattened']}
                        for r in row_names]
        return table, row_metadata, column_names

    @staticmethod
    def get_table_data_artifact(results):
        row_names = [
            'cartpole',
            'tworooms-noneuler-latest',
            'helicopter',
            'cruise-latest',
            'dcdc',
            '10rooms',
            'truck_trailer',
            'traffic_30m',
            'vehicle',
            'aircraft'
        ]
        column_names = ['CART',
                        'LinearClassifierDT-LinearSVC',
                        'LinearClassifierDT-LogisticRegression',
                        'OC1',
                        'MaxFreqDT',
                        'MaxFreq-LinearClassifierDT-LogisticRegression',
                        'MinNormDT',
                        'MinNorm-LinearClassifierDT',
                        ]
        table = []
        for dataset in row_names:
            row = []
            for classifier in column_names:
                try:
                    cell = results[dataset]['classifiers'][classifier]
                    cell['stats']['paths'] = int((cell['stats']['nodes'] + 1) / 2)
                    if 'bandwidth' in cell['stats']:
                        del cell['stats']['bandwidth']
                    if 'nodes' in cell['stats']:
                        del cell['stats']['nodes']
                    if 'oblique' in cell['stats']:
                        del cell['stats']['oblique']
                    if 'extended' in cell['stats']:
                        del cell['stats']['extended']
                except (KeyError, TypeError):
                    cell = 'not yet computed'
                row.append(cell)
            table.append(row)
        row_metadata = [{'name': r, 'domain_of_controller': results[r]['metadata']['Y_metadata'][
            'num_rows'] if r in results else "unknown",
                         'state_action_pairs': results[r]['metadata']['Y_metadata'][
                             'num_flattened'] if r in results else "unknown"} for r in row_names]
        return table, row_metadata, column_names

    def get_dot_and_c_links(self, row_metadata, column_names):
        links_table = []
        for i in range(len(row_metadata)):
            dataset = row_metadata[i]["name"]
            l = []
            for j in range(len(column_names)):
                classifier = column_names[j]
                d = {}
                path = join(self.output_folder, classifier, dataset, classifier)
                if exists(path + '.dot'):
                    d['dot_link'] = self.get_dot_link(path + '.dot')
                if exists(path + '.c'):
                    d['c_link'] = self.get_file_link(path + '.c')
                l.append(d)
            links_table.append(l)
        return links_table

    def get_dot_link(self, file):
        if getsize(file) > 500e3:
            return self.get_file_link(file)
        with open(file) as infile:
            dot = infile.read()
        return GRAPHVIZ_URL + quote(dot)

    @staticmethod
    def get_file_link(file):
        return f'file://{abspath(file)}'



from abc import ABC

from dtcontrol.decision_tree.impurity.impurity_measure import ImpurityMeasure

class MultiLabelImpurityMeasure(ImpurityMeasure, ABC):

    def get_oc1_name(self):
        return None



import copy
import sys

import numpy as np

from dtcontrol.decision_tree.splitting.split import Split
from dtcontrol.decision_tree.splitting.splitting_strategy import SplittingStrategy

class CategoricalMultiSplittingStrategy(SplittingStrategy):
    def __init__(self, value_grouping=False, tolerance=1e-5):
        """
        Implements splitting on a single categorical feature, with possibly multiple branches.
        :param value_grouping: if True, tries to merge different branches using the attribute value grouping heuristic
        :param tolerance: the absolute increase in impurity measure a value grouping may produce in order to still be
        considered a better candidate than the original (non-grouped) split
        """
        self.value_grouping = value_grouping
        self.tolerance = tolerance

    def find_split(self, dataset, impurity_measure):
        x_categorical = dataset.get_categorical_x()
        splits = {}
        for feature in range(x_categorical.shape[1]):
            real_feature = dataset.map_categorical_feature_back(feature)
            split = CategoricalMultiSplit(real_feature)
            impurity = impurity_measure.calculate_impurity(dataset, split)
            if impurity == sys.maxsize:
                continue

            if self.value_grouping:
                value_groups, grouped_impurity = \
                    self.find_best_value_groups(dataset, impurity_measure, feature, impurity)
                grouped_split = CategoricalMultiSplit(real_feature, value_groups)
                splits[grouped_split] = grouped_impurity
            else:
                splits[split] = impurity

        if not splits:
            return None
        return min(splits.keys(), key=splits.get)

    def find_best_value_groups(self, dataset, impurity_measure, feature, initial_impurity):
        impurity = initial_impurity
        real_feature = dataset.map_categorical_feature_back(feature)

        values = sorted(set(dataset.get_categorical_x()[:, feature]))
        value_groups = [[v] for v in values]
        best_new_value_groups = value_groups
        best_new_impurity = impurity
        while best_new_impurity <= impurity or abs(best_new_impurity - impurity) <= self.tolerance:
            impurity = best_new_impurity
            value_groups = best_new_value_groups
            if len(value_groups) == 2:
                break
            best_new_impurity = sys.maxsize
            for i in range(len(value_groups)):
                for j in range(i + 1, len(value_groups)):
                    new_groups = copy.deepcopy(value_groups)
                    new_groups[i] += new_groups[j]
                    del new_groups[j]
                    new_split = CategoricalMultiSplit(real_feature, new_groups)
                    new_impurity = impurity_measure.calculate_impurity(dataset, new_split)
                    if new_impurity <= best_new_impurity:
                        best_new_impurity = new_impurity
                        best_new_value_groups = new_groups
        return value_groups, impurity

class CategoricalMultiSplit(Split):
    def __init__(self, feature, value_groups=None):
        self.feature = feature
        self.value_groups = value_groups
        if not self.value_groups:
            self.value_groups = []

    def predict(self, features):
        v = features[:, self.feature][0]
        for i in range(len(self.value_groups)):
            if v in self.value_groups[i]:
                return i
        assert False

    def get_masks(self, dataset):
        if not self.value_groups:
            self.value_groups = [[v] for v in sorted(set(dataset.x[:, self.feature]))]
        masks = []
        for group in self.value_groups:
            mask = np.zeros(len(dataset.x), dtype=bool)
            for v in group:
                mask |= dataset.x[:, self.feature] == v
            masks.append(mask)
        return masks

    def print_dot(self, variables=None, category_names=None):
        if variables:
            return variables[self.feature]
        return f'x[{self.feature}]'

    def print_c(self):
        return f'x[{self.feature}]'

    def print_vhdl(self):
        return f'x{self.feature}]'



import sys

import numpy as np

from dtcontrol.decision_tree.impurity.deterministic_impurity_measure import DeterministicImpurityMeasure

class Entropy(DeterministicImpurityMeasure):
    def calculate_impurity(self, dataset, split):
        if len(split.get_masks(dataset)) == 1:
            return sys.maxsize
        impurity = 0
        for mask in split.get_masks(dataset):
            subset_labels = self.determinizer.determinize(dataset.from_mask_optimized(mask))
            if len(subset_labels) == 0:
                return sys.maxsize
            impurity += (len(subset_labels) / len(dataset)) * self.calculate_entropy(subset_labels)
        return impurity

    @staticmethod
    def calculate_entropy(y):
        num_labels = len(y)
        unique = np.unique(y)
        probabilities = [len(y[y == label]) / num_labels for label in unique]
        return sum(-prob * np.log2(prob) for prob in probabilities)

    def get_oc1_name(self):
        return 'info_gain'



from operator import itemgetter

import numpy as np

from dtcontrol.decision_tree.determinization.determinizer import Determinizer

class MaxFreqMultiDeterminizer(Determinizer):
    """
    WARNING: this determinizer is no longer supported!

    This determinizer uses the maximum frequency multi determinization approach.
    """

    def __init__(self, pre_determinize=True):
        super().__init__()
        self.pre_determinize = pre_determinize

    def determinize(self, dataset):
        """
        Given a y_train such as
        array([[[ 1,  2,  3],
                [ 1,  2,  1],
                [ 1,  2,  2],
                [ 3,  3, -1]],

               [[ 3,  4,  5],
                [ 3,  4,  4],
                [ 2,  6,  1],
                [ 3,  6, -1]]])

        gets determinized to

        array([[[ 1, -1, -1],
                [ 1, -1, -1],
                [ 1, -1, -1],
                [ 3, -1, -1]],

               [[ 3, -1, -1],
                [ 3, -1, -1],
                [ 2, -1, -1],
                [ 3, -1, -1]]])

        which is reduced to

        array([[[1],
                [1],
                [1],
                [3]],

               [[3],
                [3],
                [2],
                [3]]])
        """
        y = np.copy(dataset.y)
        determinized = False

        # list of tuples (ctrl_input_idx, input_encoding) which were already considered for keeping
        already_considered = set()

        while not determinized:
            ranks = self.get_ranks(y)

            # find the ctrl_idx and inp_enc which should be used in the next round of pruning
            # i.e. the first one from the ranking list whose input has not been already considered
            ctrl_idx = None
            inp_enc = None
            for (ctr, inp, _) in ranks:
                if (ctr, inp) not in already_considered:
                    already_considered.add((ctr, inp))
                    ctrl_idx = ctr
                    inp_enc = inp
                    break

            # Go through y[ctrl_idx] row by row
            # for each row, if it contains input_encoding, then change the remaining into -1
            # make the same -1 changes for rest of the control inputs
            for i in range(y.shape[1]):
                row = y[ctrl_idx, i]
                if inp_enc in row:
                    for j in range(y.shape[2]):
                        if row[j] != inp_enc:
                            y[:, i, j] = -1

            # check if all rows contain only one element
            determinized = True
            for ctrl_idx in range(y.shape[0]):
                for i in range(y.shape[1]):
                    row = y[ctrl_idx, i]
                    valid_row = row[row != -1]
                    determinized = determinized & (valid_row.size == 1)
        valid_y = np.array([np.array([yyy[yyy != -1] for yyy in yy]) for yy in y])
        zipped = np.stack(valid_y, axis=2)
        # [[[1,3], [1,3], [1,2], [3,3]]] -> [1,1,2,3] (tuple ids)
        return np.apply_along_axis(lambda x: self.dataset.get_tuple_to_tuple_id()[tuple(x)], axis=2,
                                   arr=zipped).flatten()

    def get_index_label(self, label):
        return self.dataset.map_tuple_id_back(label)

    def determinize_once_before_construction(self):
        return False

    def is_only_multioutput(self):
        return True

    def is_pre_split(self):
        return self.pre_determinize

    @staticmethod
    def get_ranks(y):
        """
        Generate a list of tuples (ctrl_idx, inp_enc, freq)
        where ctrl_idx is the control input index, inp_enc is the control input integer encoding and
        freq is the number of times the respective control input has occurred as the ctrl_idx'th component
        """
        ranks = []
        for ctrl_idx in range(y.shape[0]):
            flattened_control = y[ctrl_idx].flatten()
            flattened_control = flattened_control[flattened_control != -1]
            counter = list(zip(range(len(np.bincount(flattened_control))), np.bincount(flattened_control)))
            idx_input_count = [(ctrl_idx,) + l for l in counter]
            ranks.extend(idx_input_count)
        return sorted(ranks, key=itemgetter(2), reverse=True)



import unittest

from dtcontrol.dataset.single_output_dataset import SingleOutputDataset
from dtcontrol.decision_tree.decision_tree import DecisionTree
from dtcontrol.decision_tree.impurity.entropy import Entropy
from dtcontrol.decision_tree.splitting.axis_aligned import AxisAlignedSplittingStrategy
from dtcontrol.decision_tree.splitting.categorical_multi import CategoricalMultiSplittingStrategy

class TestCategorical(unittest.TestCase):
    def test_categorical(self):
        try:
            ds = SingleOutputDataset('golf.csv')
            ds.load_if_necessary()
        except FileNotFoundError:
            ds = SingleOutputDataset('test/golf.csv')
            ds.load_if_necessary()
        self.assertEqual([0, 1, 2, 3], ds.x_metadata['categorical'])
        self.assertEqual(["Weather", "Temperature", "Humidity", "Wind"], ds.x_metadata['variables'])
        self.assertEqual({
            0: ["Rain", "Overcast", "Sunny"],
            1: ["Cool", "Mild", "Hot"],
            3: ["Weak", "Strong"]
        }, ds.x_metadata['category_names'])

        dt = DecisionTree([CategoricalMultiSplittingStrategy(), AxisAlignedSplittingStrategy()],
                          Entropy(), 'categorical')
        dt.fit(ds)
        root = dt.root
        self.assertEqual(0, root.split.feature)
        self.assertEqual(3, len(root.children))
        l = root.children[0]
        self.assertTrue(l.is_leaf())
        self.assertEqual(0, l.actual_label)
        m = root.children[1]
        self.assertEqual(1, m.split.feature)
        self.assertEqual(3, len(m.children))
        ml = m.children[0]
        self.assertTrue(ml.is_leaf())
        self.assertEqual(1, ml.actual_label)
        mm = m.children[0]
        self.assertTrue(mm.is_leaf())
        self.assertEqual(1, mm.actual_label)
        mr = m.children[2]
        self.assertTrue(mr.is_leaf())
        self.assertEqual(0, mr.actual_label)
        r = root.children[2]
        self.assertEqual(3, r.split.feature)
        self.assertEqual(2, len(r.children))
        rl = r.children[0]
        rr = r.children[1]
        self.assertTrue(rl.is_leaf())
        self.assertEqual(1, rl.actual_label)
        self.assertTrue(rr.is_leaf())
        self.assertEqual(0, rr.actual_label)



from abc import ABC, abstractmethod

class Split(ABC):
    @abstractmethod
    def predict(self, features):
        """
        Determines the child index of the split for one particular instance.
        :param features: the features of the instance
        :returns: the child index (0/1 for a binary split)
        """
        pass

    def split(self, dataset):
        """
        Splits the dataset into subsets.
        :param dataset: the dataset to be split
        :return: a list of the subsets
        """
        return [dataset.from_mask(mask) for mask in self.get_masks(dataset)]

    @abstractmethod
    def get_masks(self, dataset):
        """
        Returns the masks specifying this split.
        :param dataset: the dataset to be split
        :return: a list of the masks corresponding to each subset after the split
        """

    @abstractmethod
    def print_dot(self, variables=None, category_names=None):
        pass

    @abstractmethod
    def print_c(self):
        pass

    @abstractmethod
    def print_vhdl(self):
        pass



from abc import ABC

import numpy as np

from dtcontrol.decision_tree.splitting.split import Split

class LinearSplit(Split, ABC):
    """
    Represents a linear split of the form wTx + b <= 0.
    """

    def __init__(self, coefficients, intercept, real_coefficients, numeric_columns):
        """
        :param coefficients: the coefficients with respect to the numeric columns
        :param intercept: the intercept
        :param real_coefficients: the coefficients with respect to all columns
        :param numeric_columns: a list of integers defining the numeric columns
        """
        self.coefficients = coefficients
        self.intercept = intercept
        self.real_coefficients = real_coefficients
        self.numeric_columns = numeric_columns

    @staticmethod
    def map_numeric_coefficients_back(numeric_coefficients, dataset):
        dim = dataset.x.shape[1]
        new_coefficients = [0] * dim
        for i in range(len(numeric_coefficients)):
            new_coefficients[dataset.map_numeric_feature_back(i)] = numeric_coefficients[i]
        return np.array(new_coefficients)

    def get_masks(self, dataset):
        mask = np.dot(dataset.get_numeric_x(), self.coefficients) + self.intercept <= 0
        return [mask, ~mask]

    def predict(self, features):
        return 0 if np.dot(features[:, self.numeric_columns], self.coefficients) + self.intercept <= 0 else 1

    def print_dot(self, variables=None, category_names=None):
        return self.get_hyperplane_str(rounded=True, newlines=True, variables=variables)

    def print_c(self):
        return self.get_hyperplane_str()

    def print_vhdl(self):
        hyperplane = self.get_hyperplane_str()
        hyperplane.replace('[', '')
        hyperplane.replace(']', '')
        return hyperplane

    def get_hyperplane_str(self, rounded=False, newlines=False, variables=None):
        line = []
        for i in range(len(self.real_coefficients)):
            if self.real_coefficients[i] == 0:
                continue
            coefficient = round(self.real_coefficients[i], 6) if rounded else self.real_coefficients[i]
            variable = variables[i] if variables else f'x[{i}]'
            line.append(f"{coefficient}*{variable}")
        line.append(f"{round(self.intercept, 6) if rounded else self.intercept}")
        joiner = "\\n+" if newlines else "+"
        hyperplane = joiner.join(line) + " <= 0"
        return hyperplane.replace('+-', '-')



import json
import logging
import pickle
from abc import ABC, abstractmethod
from os import makedirs
from os.path import getmtime, split, exists, join, splitext, isfile

import numpy as np

import dtcontrol.util as util
from dtcontrol.util import is_int

class DatasetLoader(ABC):
    def __init__(self):
        self.loaded_datasets = {}
        self.PATH = '.benchmark_suite'

    def load_dataset(self, filename):
        if filename not in self.loaded_datasets:
            if self.is_already_converted(filename):
                self.loaded_datasets[filename] = self.load_converted_dataset(filename)
            else:
                self.loaded_datasets[filename] = self.load_dataset_and_config(filename)
                self.save_converted_dataset(filename)
        return self.loaded_datasets[filename]

    def is_already_converted(self, filename):
        possible_path = self.get_converted_folder(filename)
        return exists(possible_path)

    def load_converted_dataset(self, filename):
        folder = self.get_converted_folder(filename)
        assert exists(folder)
        util.log_without_newline('Loading existing converted dataset...')
        x = np.load(join(folder, 'X_train.npy'))
        y = np.load(join(folder, 'Y_train.npy'))
        with open(join(folder, 'extra_data.pickle'), 'rb') as infile:
            extra_data = pickle.load(infile)
            index_to_actual = extra_data["index_to_value"]
            x_metadata = extra_data["X_metadata"]
            y_metadata = extra_data["Y_metadata"]
        logging.info(" Done.")
        return x, x_metadata, y, y_metadata, index_to_actual

    def save_converted_dataset(self, filename):
        folder = self.get_converted_folder(filename)
        if not exists(folder):
            makedirs(folder)
        x, x_metadata, y, y_metadata, index_to_actual = self.loaded_datasets[filename]
        np.save(join(folder, 'X_train.npy'), x)
        np.save(join(folder, 'Y_train.npy'), y)
        with open(join(folder, 'extra_data.pickle'), 'wb+') as outfile:
            pickle.dump({"index_to_value": index_to_actual,
                         "X_metadata": x_metadata,
                         "Y_metadata": y_metadata},
                        outfile)

    def get_converted_folder(self, filename):
        directory, name = split(filename)
        return join(directory, self.PATH, f'{name}_{getmtime(filename)}')

    def load_dataset_and_config(self, filename):
        tup = self._load_dataset(filename)
        name, _ = splitext(filename)
        config_name = name + '_config.json'
        if exists(config_name) and isfile(config_name):
            self.update_with_config(config_name, tup[0], tup[1], tup[3])
        return tup

    @staticmethod
    def update_with_config(config_filename, x, x_metadata, y_metadata):
        with open(config_filename, 'r') as infile:
            config = json.load(infile)
        if 'x_column_types' in config:
            DatasetLoader.update_column_types(config['x_column_types'], x_metadata, config_filename)
        if 'y_column_types' in config:
            DatasetLoader.update_column_types(config['y_column_types'], y_metadata, config_filename)

        if 'x_column_names' in config:
            column_names = config['x_column_names']
            if len(column_names) != x.shape[1]:
                raise ValueError(f'\'{config_filename}\': the number of column names '
                                 f'does not match the actual number of columns.')
            x_metadata['variables'] = column_names

        if 'x_category_names' in config:
            DatasetLoader.update_category_names(config['x_category_names'], x_metadata, config_filename)
        if 'y_category_names' in config:
            DatasetLoader.update_category_names(config['y_category_names'], y_metadata, config_filename)

    @staticmethod
    def update_column_types(column_types, metadata, config_filename):
        if 'categorical' in column_types:
            metadata['categorical'] = column_types['categorical']
        if 'numeric' in column_types:
            # columns are numeric by default, we only need to check for conflicts here
            if len(set(column_types['numeric']).intersection(column_types['categorical'])) != 0:
                raise ValueError(f'\'{config_filename}\' contains columns marked as '
                                 f'categorical as well as numeric.')

    @staticmethod
    def update_category_names(category_names, metadata, config_filename):
        column_to_index = {}
        for column in category_names:
            if not is_int(column):
                try:
                    column_to_index[column] = metadata['variables'].index(column)
                except ValueError:
                    raise ValueError(f'\'{config_filename}\': unknown column name in category_names.')
            else:
                column_to_index[column] = int(column)
            if column_to_index[column] not in metadata['categorical']:
                raise ValueError(f'\'{config_filename}\': numeric column in category_names.')
        metadata['category_names'] = {column_to_index[k]: v for k, v in category_names.items()}

    @abstractmethod
    def _load_dataset(self, filename):
        """
        Loads a dataset and returns x, x_metadata, y, y_metadata, index_to_actual
        """
        pass



from dtcontrol.decision_tree.splitting.split import Split
from dtcontrol.decision_tree.splitting.splitting_strategy import SplittingStrategy

class AxisAlignedSplittingStrategy(SplittingStrategy):
    def find_split(self, dataset, impurity_measure):
        x_numeric = dataset.get_numeric_x()
        splits = {}
        for feature in range(x_numeric.shape[1]):
            values = sorted(set(x_numeric[:, feature]))
            for i in range(len(values) - 1):
                threshold = (values[i] + values[i + 1]) / 2
                real_feature = dataset.map_numeric_feature_back(feature)
                split = AxisAlignedSplit(real_feature, threshold)
                splits[split] = impurity_measure.calculate_impurity(dataset, split)
        if not splits:
            return None
        return min(splits.keys(), key=splits.get)

class AxisAlignedSplit(Split):
    """
    Represents an axis aligned split of the form x[i] <= b.
    """

    def __init__(self, feature, threshold):
        self.feature = feature
        self.threshold = threshold

    def get_masks(self, dataset):
        mask = dataset.x[:, self.feature] <= self.threshold
        return [mask, ~mask]

    def predict(self, features):
        return 0 if features[:, self.feature][0] <= self.threshold else 1

    def print_dot(self, variables=None, category_names=None):
        if variables:
            return f'{variables[self.feature]} <= {round(self.threshold, 6)}'
        return self.print_c()

    def print_c(self):
        return f'x[{self.feature}] <= {round(self.threshold, 6)}'

    def print_vhdl(self):
        return f'x{self.feature} <= {round(self.threshold, 6)}'



from abc import ABC, abstractmethod

class Determinizer(ABC):
    def __init__(self):
        self.pre_determinized_labels = None

    @abstractmethod
    def determinize(self, dataset):
        """
        :param dataset: the dataset to be determinized
        :return: the determinized labels
        """
        pass

    @abstractmethod
    def is_pre_split(self):
        """
        Returns true if this determinizer should use the pre-split optimization.
        """
        pass



import glob
import json
import logging
import os
import sys
import time
import traceback
import webbrowser
from os import makedirs
from os.path import join, exists, isfile

import numpy as np
from jinja2 import FileSystemLoader, Environment

import dtcontrol
from dtcontrol import util
from dtcontrol.dataset.multi_output_dataset import MultiOutputDataset
from dtcontrol.dataset.single_output_dataset import SingleOutputDataset
from dtcontrol.timeout import call_with_timeout
from dtcontrol.ui.table_controller import TableController
from dtcontrol.util import format_seconds, get_filename_and_ext

file_loader = FileSystemLoader([path + "/c_templates" for path in dtcontrol.__path__])
env = Environment(loader=file_loader)
single_output_c_template = env.get_template('single_output.c')
multi_output_c_template = env.get_template('multi_output.c')

util.ignore_convergence_warnings()

class BenchmarkSuite:
    """
    The benchmark suite runs the given classifiers on the given datasets and saves the results.
    """

    def __init__(self, benchmark_file='benchmark', timeout=None, output_folder='decision_trees', save_folder=None,
                 rerun=False, is_artifact=False):
        logging.basicConfig(level=logging.INFO, format='%(message)s', stream=sys.stdout)
        self.datasets = []
        self.json_file = f'{benchmark_file}.json'
        self.html_file = f'{benchmark_file}.html'
        self.results = {}
        self.timeout = timeout
        self.output_folder = output_folder
        self.save_folder = save_folder
        self.rerun = rerun  # rerun benchmarks even if there already are results available
        self.is_artifact = is_artifact  # always produces a table exactly corresponding to the one in the paper
        self.table_controller = TableController(self.html_file, self.output_folder, self.is_artifact)

        logging.info(f"INFO: Benchmark statistics will be available in {self.json_file} and {self.html_file}.")
        logging.info(f"INFO: Constructed trees will be written to {self.output_folder}.\n")

    def add_datasets(self, paths, include=None, exclude=None):
        if not exclude:
            exclude = []
        if include is not None and len(set(include) & set(exclude)) > 0:
            logging.error('A dataset cannot be both included and excluded.\nAborting.')
            return
        self.datasets = []
        if isinstance(paths, str):
            paths = [paths]
        for path in paths:
            for file in self.get_files(path):
                name, ext = get_filename_and_ext(file)
                if ((not include) or name in include) and name not in exclude:
                    if self.is_multiout(file, ext):
                        ds = MultiOutputDataset(file)
                    else:
                        ds = SingleOutputDataset(file)

                    ds.is_deterministic = self.is_deterministic(file, ext)
                    self.datasets.append(ds)
        self.datasets.sort(key=lambda ds: ds.get_name())

    @staticmethod
    def get_files(path):
        if isfile(path):
            return [path]
        else:
            return [p for ext in ['*.scs', '*.dump', '*.csv', '*.prism'] for p in glob.glob(join(path, ext))
                    if not p.endswith('_states.prism')]

    def display_html(self):
        url = f'file://{os.path.abspath(self.html_file)}'
        webbrowser.open(url)

    def benchmark(self, classifiers):
        self.load_results()
        num_steps = len(classifiers) * len(self.datasets)
        step = 0
        for ds in self.datasets:
            for classifier in classifiers:
                step += 1
                logging.info(f"{step}/{num_steps}: Evaluating {classifier.get_name()} on {ds.get_name()}... ")
                try:
                    cell, computed = self.compute_cell(ds, classifier)
                except ValueError as e:
                    # traceback.print_exc()
                    logging.error(e)
                    continue
                if computed:
                    self.save_result(classifier.get_name(), ds, cell)
                    if cell == 'timeout':
                        msg = f"{step}/{num_steps}: Timed out after {format_seconds(self.timeout)}"
                    elif cell == 'failed to fit':
                        msg = f"{step}/{num_steps}: Failed to fit"
                    else:
                        msg = f"{step}/{num_steps}: Finished in {cell['time']}."
                    logging.info(msg)
                else:
                    if cell == 'not applicable':
                        self.save_result(classifier.get_name(), ds, cell)
                        logging.info(
                            f"{step}/{num_steps}: Not applicable.")
                    else:
                        logging.info(
                            f"{step}/{num_steps}: Not running since the result is already available.")
        logging.info('All benchmarks completed. Shutting down dtControl.')

        self.table_controller.update_and_save(self.results, [ds.get_name() for ds in self.datasets],
                                              [cl.get_name() for cl in classifiers])

    def compute_cell(self, dataset, classifier):
        if self.already_computed(dataset, classifier) and not self.rerun:
            computed = False
            cell = self.results[dataset.get_name()]['classifiers'][classifier.get_name()]
        elif not classifier.is_applicable(dataset):
            computed = False
            cell = 'not applicable'
        else:
            computed = True
            dataset.load_if_necessary()
            cell = self.train_and_get_cell(dataset, classifier)
        return cell, computed

    def already_computed(self, dataset, classifier):
        return dataset.get_name() in self.results and classifier.get_name() in self.results[dataset.get_name()][
            'classifiers']

    def train_and_get_cell(self, dataset, classifier):
        if self.timeout is not None:
            classifier, success, run_time = call_with_timeout(classifier, 'fit', dataset, timeout=self.timeout)
        else:
            start = time.time()
            classifier.fit(dataset)
            run_time = time.time() - start
            success = True
        if success:
            # todo once BDD is working for all models
            # if isinstance(classifier, BDD):  # bdd cannot predict yet
            #     stats = classifier.get_stats()
            #     cell = {'stats': stats, 'time': format_seconds(run_time)}
            #     return cell
            acc = dataset.compute_accuracy(classifier.predict(dataset, actual_values=False))
            if acc is None:
                cell = 'failed to fit'
            else:
                stats = classifier.get_stats()
                cell = {'stats': stats, 'time': format_seconds(run_time)}
                self.save_dot_and_c(classifier, dataset)
                # if not isinstance(classifier, OC1Wrapper):
                #     vhdl_filename = self.get_filename(self.output_folder, dataset, classifier, '.vhdl')
                #     classifier.print_vhdl(len(dataset.x_metadata["variables"]), vhdl_filename)
                if abs(acc - 1.0) > 1e-10:
                    cell['accuracy'] = acc
                if self.save_folder is not None:
                    classifier.save(self.get_filename(self.save_folder, dataset, classifier, '.saved', unique=True))
        else:
            cell = 'timeout'
        return cell

    def save_dot_and_c(self, classifier, dataset):
        dot_filename = self.get_filename(self.output_folder, dataset, classifier, '.dot')
        with open(dot_filename, 'w+') as outfile:
            outfile.write(classifier.print_dot(dataset.x_metadata, dataset.y_metadata))

        num_outputs = 1 if len(dataset.y.shape) <= 2 else len(dataset.y)
        template = multi_output_c_template if num_outputs > 1 else single_output_c_template
        example = f'{{{",".join(str(i) + (".f" if isinstance(i, np.integer) else "f") for i in dataset.x[0])}}}'
        c_filename = self.get_filename(self.output_folder, dataset, classifier, '.c')
        with open(c_filename, 'w+') as outfile:
            outfile.write(template.render(example=example, num_outputs=num_outputs, code=classifier.print_c()))

    @staticmethod
    def get_filename(folder, dataset, classifier, extension, unique=False):
        dir = join(folder, classifier.get_name(), dataset.get_name())
        if not exists(dir):
            makedirs(dir)
        name = classifier.get_name()
        if unique:
            name += f'--{time.strftime("%Y%m%d-%H%M%S")}'
        name += extension
        return join(dir, name)

    def save_result(self, classifier_name, dataset, result):
        if dataset.get_name() not in self.results:
            self.results[dataset.get_name()] = {'classifiers': {},
                                                'metadata': {
                                                    'X_metadata': dataset.x_metadata,
                                                    'Y_metadata': dataset.y_metadata
                                                }
                                                }
        self.results[dataset.get_name()]['classifiers'][classifier_name] = result
        self.save_to_disk()

    def load_results(self):
        if not self.json_file or not exists(self.json_file) or not isfile(self.json_file):
            return
        with open(self.json_file, 'r') as infile:
            self.results = json.load(infile)
        for ds in self.datasets:
            if ds.get_name() in self.results:
                ds.load_metadata_from_json(self.results[ds.get_name()])

    def delete_dataset_results(self, dataset_name):
        self.load_results()
        if dataset_name in self.results:
            del self.results[dataset_name]
        self.save_to_disk()

    def delete_classifier_results(self, classifier_name):
        self.load_results()
        for dataset in self.results:
            classifiers = self.results[dataset]['classifiers']
            if classifier_name in classifiers:
                del classifiers[classifier_name]
        self.save_to_disk()

    def delete_cell(self, dataset_name, classifier_name):
        self.load_results()
        if dataset_name in self.results:
            classifiers = self.results[dataset_name]['classifiers']
            if classifier_name in classifiers:
                del classifiers[classifier_name]
        self.save_to_disk()

    def save_to_disk(self):
        with open(self.json_file, 'w+') as outfile:
            json.dump(self.results, outfile, indent=2)

    @staticmethod
    def is_multiout(filename, ext):
        if "scs" in ext:
            # if scs, then
            with open(filename) as f:
                # Read input dim from scs file
                for i in range(5):
                    f.readline()
                state_dim = int(f.readline())
                for i in range(12 + 3 * state_dim):
                    f.readline()
                input_dim = int(f.readline())
                return input_dim > 1
        elif "csv" in ext:
            with open(filename) as f:
                f.readline()
                _, input_dim = map(int, f.readline().split("BEGIN")[1].split())
                return input_dim > 1
        else:
            return False

    @staticmethod
    def is_deterministic(filename, ext):
        if "scs" in ext:
            with open(filename) as f:
                # Read input dim from scs file
                for i in range(5):
                    f.readline()
                state_dim = int(f.readline())
                for i in range(12 + 3 * state_dim):
                    f.readline()
                input_dim = int(f.readline())
                for i in range(12 + 3 * input_dim):
                    f.readline()
                non_det = int(f.readline().split(":")[1].split()[1])
                return non_det == 1
        elif "csv" in ext:
            with open(filename) as f:
                is_det = "NON-PERMISSIVE" in f.readline()
                return is_det
        elif "prism" in ext:
            return True  # PRISM is always deterministic (?)
        else:
            return False  # UPPAAL is always non-deterministic



import numpy as np

from dtcontrol.decision_tree.determinization.non_determinizer import NonDeterminizer
from dtcontrol.decision_tree.splitting.linear_classifier import LinearClassifierSplit
from dtcontrol.decision_tree.splitting.linear_split import LinearSplit
from dtcontrol.decision_tree.splitting.splitting_strategy import SplittingStrategy

class LinearClassifierOnlyLeafSplittingStrategy(SplittingStrategy):
    def __init__(self, classifier_class, determinizer=NonDeterminizer(), **kwargs):
        self.determinizer = determinizer
        self.classifier_class = classifier_class
        self.kwargs = kwargs

    def find_split(self, dataset, impurity_measure):
        x_numeric = dataset.get_numeric_x()
        if x_numeric.shape[1] == 0:
            return None
        y = self.determinizer.determinize(dataset)
        if not self.is_binary(y):  # otherwise we certainly won't have pure leaves
            return None

        label = y[0]
        new_y = np.copy(y)
        label_mask = (new_y == label)
        new_y[label_mask] = 1
        new_y[~label_mask] = -1
        classifier = self.classifier_class(**self.kwargs)
        classifier.fit(x_numeric, new_y)

        if np.array_equal(classifier.predict(x_numeric), new_y):  # perfect split
            real_features = LinearSplit.map_numeric_coefficients_back(classifier.coef_[0], dataset)
            split = LinearClassifierSplit(classifier, real_features, dataset.numeric_columns)
            assert impurity_measure.calculate_impurity(dataset, split) == 0
            return split
        return None

    @staticmethod
    def is_binary(y):
        return len(np.unique(y)) == 2



from dtcontrol.decision_tree.impurity.multi_label_impurity_measure import MultiLabelImpurityMeasure
from dtcontrol.decision_tree.impurity.scaled_bincount import ScaledBincount

class MultiLabelGiniIndex(MultiLabelImpurityMeasure):

    def __init__(self):
        self.scaled_bincount = ScaledBincount(self.scaling_function)

    @staticmethod
    def scaling_function(x):
        return x**2

    def calculate_impurity(self, dataset, split):
        return self.scaled_bincount.calculate_impurity(dataset, split)



import logging

import numpy as np
from tqdm import tqdm

from dtcontrol.dataset.dataset_loader import DatasetLoader

class ScotsDatasetLoader(DatasetLoader):
    def _load_dataset(self, filename):
        precision = 10

        with open(filename) as f:
            logging.info(f"Reading from {filename}")

            for i in range(5):
                f.readline()

            line = f.readline()
            state_dim = int(line)

            for i in range(2):
                f.readline()

            state_eta = []
            for i in range(state_dim):
                line = f.readline()
                state_eta.append(round(float(line), precision))

            for i in range(3):
                f.readline()

            state_lb = []
            for i in range(state_dim):
                line = f.readline()
                state_lb.append(round(float(line), precision))

            for i in range(3):
                f.readline()

            state_ub = []
            for i in range(state_dim):
                line = f.readline()
                state_ub.append(round(float(line), precision))

            n_state_grid = []
            for i in range(state_dim):
                n_state_grid.append(int(round((state_ub[i] - state_lb[i]) / state_eta[i], precision)) + 1)

            for i in range(4):
                f.readline()

            input_dim = int(f.readline())

            for i in range(2):
                f.readline()

            input_eta = []
            for i in range(input_dim):
                line = f.readline()
                input_eta.append(round(float(line), precision))

            for i in range(3):
                f.readline()

            input_lb = []
            for i in range(input_dim):
                line = f.readline()
                input_lb.append(round(float(line), precision))

            for i in range(3):
                f.readline()

            input_ub = []
            for i in range(input_dim):
                line = f.readline()
                input_ub.append(float(line))

            n_input_grid = []
            for i in range(input_dim):
                n_input_grid.append(int(round((input_ub[i] - input_lb[i]) / input_eta[i], precision) + 1))

            for i in range(4):
                f.readline()

            dim_str = f.readline().split(":")[1]
            rows, max_non_det = list(map(int, dim_str.split()))

            # Controller starts now
            controller_start = 32 + 3 * (state_dim + input_dim)

            x_nn = [1]
            for i in range(1, state_dim):
                x_nn.append(x_nn[i - 1] * n_state_grid[i - 1])

            u_nn = [1]
            for i in range(1, input_dim):
                u_nn.append(u_nn[i - 1] * n_input_grid[i - 1])

            # Get the number of lines describing the controller
            controller_lines = sum(1 for line in f) - 1

            x = np.empty((controller_lines, state_dim), dtype=np.float32)
            y = np.full((input_dim, controller_lines, max_non_det), -1, dtype=np.int32)

            f.seek(0)

            float_to_unique_label = dict()

            for i in range(controller_start):
                f.readline()

            logging.info("Extracting states and control inputs from SCOTS dump")
            for i, line in enumerate(tqdm(f, total=controller_lines)):
                if i == controller_lines:
                    break
                idxu = np.fromstring(line, dtype=np.int32, sep=' ')
                idx = idxu[0]

                k = state_dim - 1
                x2 = np.zeros(state_dim)
                while k > 0:
                    num = int(idx / x_nn[k])  # j
                    idx = idx % x_nn[k]  # i
                    x2[k] = state_lb[k] + num * state_eta[k]
                    k = k - 1
                num = idx
                x2[0] = state_lb[0] + num * state_eta[0]

                x[i] = x2

                # creating input variables
                u_idx = np.empty(input_dim, dtype=np.int16)
                for j in range(1, idxu.shape[0]):
                    idu = idxu[j]
                    kk = input_dim - 1
                    u = np.empty(input_dim, dtype=np.float32)
                    while kk > 0:
                        u_idx[kk] = int(idu / u_nn[kk])
                        idu = idu % u_nn[kk]
                        u[kk] = input_lb[kk] + u_idx[kk] * input_eta[kk]
                        if u[kk] not in float_to_unique_label.keys():
                            float_to_unique_label[u[kk]] = len(float_to_unique_label) + 1
                        y[kk][i][j - 1] = float_to_unique_label[u[kk]]
                        kk = kk - 1
                    u_idx[0] = idu
                    u[0] = input_lb[0] + u_idx[0] * input_eta[0]
                    if u[0] not in float_to_unique_label.keys():
                        float_to_unique_label[u[0]] = len(float_to_unique_label) + 1
                    y[0][i][j - 1] = float_to_unique_label[u[0]]

            # inverse map
            unique_label_to_float = {y: x for (x, y) in float_to_unique_label.items()}

            # if only single control input, do not wrap it in another array
            if y.shape[0] == 1:
                y = y[0]

            logging.info("Constructed training set with %s datapoints" % x.shape[0])

            # construct metadata
            x_metadata = dict()
            x_metadata["categorical"] = []
            x_metadata["min"] = state_lb
            x_metadata["max"] = state_ub
            x_metadata["step_size"] = state_eta

            y_metadata = dict()
            y_metadata["variables"] = [f"u_{i}" for i in range(len(input_lb))]
            y_metadata["min"] = input_lb
            y_metadata["max"] = input_ub
            y_metadata["step_size"] = input_eta

            return (x, x_metadata, y, y_metadata, unique_label_to_float)



import numpy as np
from dataset.multi_output_dataset import MultiOutputDataset
from src.decision_tree.cart_custom_dt import CartNode

class MaxCartForest():
    def __init__(self):
        self.trees = None
        self.name = 'MaxCARTForest'

    def is_applicable(self, dataset):
        return isinstance(dataset, MultiOutputDataset) and not dataset.is_deterministic

    def fit(self, dataset):  # TODO: extend to do max at every node (currently only at root node)
        if dataset.tuple_to_tuple_id is None:
            dataset.get_tuple_ids()
        Y_det = dataset.determinize_max_over_all_inputs(dataset.Y_train, dataset.tuple_to_tuple_id)
        Y_det = Y_det.reshape(-1, 1)
        Y_det_tuples = np.apply_along_axis(lambda x: dataset.map_tuple_id_back(x[0]), axis=1, arr=Y_det)
        self.trees = []
        num_outputs = len(dataset.Y_train)
        for i in range(num_outputs):
            tree = CartNode()
            tree.fit(dataset.X_train, Y_det_tuples[:, i])
            tree.set_labels(lambda leaf: leaf.trained_label, dataset.index_to_value)
            self.trees.append(tree)

    def predict(self, dataset):
        return list(zip(*[self.trees[i].predict(dataset.X_train) for i in range(len(self.trees))]))

    def get_stats(self) -> dict:
        stats = {}
        for i in range(len(self.trees)):
            stats[f'num_nodes_{i}'] = self.trees[i].num_nodes
        return stats

    def save(self, file):
        pass

    def export_dot(self, file):
        pass

    def export_c(self, file):
        pass

    def __str__(self):
        return 'MaxCARTForest'



from abc import ABC, abstractmethod

from dtcontrol.benchmark_suite_classifier import BenchmarkSuiteClassifier

class PostProcessingMethod(BenchmarkSuiteClassifier, ABC):
    def __init__(self, classifier, name):
        super().__init__(name)
        self.classifier = classifier

    @abstractmethod
    def run(self):
        """
        Runs the post-processing method and transforms the decision tree.
        """
        pass

    def is_applicable(self, dataset):
        return self.classifier.is_applicable(dataset)

    def fit(self, dataset):
        self.classifier.fit(dataset)
        self.run()

    def predict(self, dataset, actual_values=True):
        return self.classifier.predict(dataset, actual_values)

    def get_stats(self):
        return self.classifier.get_stats()

    def save(self, file):
        self.classifier.save(file)

    def print_dot(self, x_metadata, y_metadata):
        return self.classifier.print_dot(x_metadata, y_metadata)

    def print_c(self):
        return self.classifier.print_c()



import numpy as np

from dtcontrol.dataset.dataset import Dataset
from dtcontrol.util import make_set

class MultiOutputDataset(Dataset):
    def __init__(self, filename):
        super().__init__(filename)
        self.tuple_ids = None
        self.tuple_id_to_tuple = None
        self.tuple_to_tuple_id = None
        self.unique_labels = None  # ids of lists of tuple ids
        self.list_id_to_list = None  # maps ids of lists of tuple ids to the actual lists of tuple ids
        # the data in a tuple (zipped) format where the outermost dimension becomes the innermost dimension
        # i.e. the new dimensions are: num_examples x num_labels x num_outputs
        self.tuples = None

    def compute_accuracy(self, y_pred):
        self.check_loaded()
        num_correct = 0
        for i in range(len(y_pred)):
            pred = y_pred[i]
            if pred is None:
                return None
            correct_tuples = set([tuple(l) for l in self.get_tuples()[i] if l[0] != -1])
            if set.issubset(make_set(pred), correct_tuples):
                num_correct += 1
        return num_correct / len(y_pred)

    def get_tuples(self):
        """
            [[[ 0, -1, -1],
              [ 0, -1, -1],
              [ 0, -1, -1],
              [ 1,  2,  0]],

             [[ 0, -1, -1],
              [ 0, -1, -1],
              [ 0, -1, -1],
              [ 0,  0,  0]]]

            gets mapped to

            [[[ 0  0]
              [-1 -1]
              [-1 -1]]

             [[ 0  0]
              [-1 -1]
              [-1 -1]]

             [[ 0  0]
              [-1 -1]
              [-1 -1]]

             [[ 1  0]
              [ 2  0]
              [ 0  0]]]
        """
        if self.tuples is None:
            self.tuples = np.stack(self.y, axis=2)
        return self.tuples

    def get_single_labels(self):
        return self.get_tuple_ids()

    def map_single_label_back(self, single_label):
        return self.map_tuple_id_back(single_label)

    def get_tuple_ids(self):
        """
        [[[ 0  0]
          [-1 -1]
          [-1 -1]]

         [[ 0  0]
          [-1 -1]
          [-1 -1]]

         [[ 0  0]
          [-1 -1]
          [-1 -1]]

         [[ 1  0]
          [ 2  0]
          [ 0  0]]]

         gets mapped to

         [[2 -1 -1]
          [2 -1 -1]
          [2 -1 -1]
          [3  4  2]]
        """
        if self.tuple_ids is not None:
            return self.tuple_ids

        stacked_y_train = self.get_tuples()

        # default
        tuple_to_index = {tuple(-1 for i in range(stacked_y_train.shape[2])): -1}

        self.tuple_ids = np.full((stacked_y_train.shape[0], stacked_y_train.shape[1]), -1)

        # first axis: datapoints
        # second axis: non-det
        # third axis: control inputs
        i = 0
        for datapoint in stacked_y_train:
            j = 0
            for action in datapoint:
                action_tuple = tuple(action)
                if action_tuple not in tuple_to_index.keys():
                    # indexing from 1
                    tuple_to_index[action_tuple] = len(tuple_to_index) + 1
                self.tuple_ids[i][j] = tuple_to_index[action_tuple]
                j += 1
            i += 1

        self.tuple_to_tuple_id = tuple_to_index
        self.tuple_id_to_tuple = {y: x for (x, y) in tuple_to_index.items()}
        return self.tuple_ids

    def get_tuple_to_tuple_id(self):
        if self.tuple_to_tuple_id is None:
            self.get_unique_labels()
        return self.tuple_to_tuple_id

    def get_unique_labels(self):
        if self.unique_labels is None:
            self.unique_labels, self.list_id_to_list = self.get_unique_labels_from_2d(self.get_tuple_ids())
        return self.unique_labels

    def map_unique_label_back(self, label):
        l = self.list_id_to_list[label]
        return [self.tuple_id_to_tuple[i] for i in l if i != -1]

    def map_tuple_id_back(self, tuple_id):
        return self.tuple_id_to_tuple[tuple_id]

    def from_mask(self, mask):
        subset = MultiOutputDataset(self.filename)
        subset.copy_from_other_dataset(self)
        subset.parent_mask = mask
        subset.x = self.x[mask]
        subset.y = self.y[:, mask, :]
        if self.tuple_ids is not None:
            subset.tuple_ids = self.tuple_ids[mask]
            subset.tuple_id_to_tuple = self.tuple_id_to_tuple
            subset.tuple_to_tuple_id = self.tuple_to_tuple_id
        if self.unique_labels is not None:
            subset.unique_labels = self.unique_labels[mask]
            subset.list_id_to_list = self.list_id_to_list
        if self.tuples is not None:
            subset.tuples = self.tuples[mask]
        return subset

    def from_mask_optimized(self, mask):
        empty_object = type('', (), {})()
        empty_object.parent_mask = mask
        empty_object.get_single_labels = lambda: self.get_single_labels()[mask]
        return empty_object



from dtcontrol.dataset.dataset import Dataset
from dtcontrol.util import make_set

class SingleOutputDataset(Dataset):
    def __init__(self, filename):
        super().__init__(filename)
        self.unique_labels = None  # a unique int for every combination of possible outputs
        self.unique_mapping = None

    def compute_accuracy(self, y_pred):
        self.check_loaded()
        num_correct = 0
        for i in range(len(y_pred)):
            pred = y_pred[i]
            if pred is None:
                return None
            if set.issubset(make_set(pred), set(self.y[i])):
                num_correct += 1
        return num_correct / len(y_pred)

    def get_single_labels(self):
        return self.y

    def map_single_label_back(self, single_label):
        return single_label

    def get_unique_labels(self):
        """
        e.g.
        [[1  2  3 -1 -1],
         [1 -1 -1 -1 -1],
         [1  2 -1 -1 -1],
        ]

        gets mapped to

        unique_labels = [1, 2, 3]
        unique_mapping = {1: [1 2 3 -1 -1], 2: [1 -1 -1 -1 -1], 3: [1 2 -1 -1 -1]}
        """
        if self.unique_labels is None:
            self.unique_labels, self.unique_mapping = self.get_unique_labels_from_2d(self.y)
        return self.unique_labels

    def map_unique_label_back(self, label):
        return [i for i in self.unique_mapping[label] if i != -1]

    def from_mask(self, mask):
        subset = SingleOutputDataset(self.filename)
        subset.copy_from_other_dataset(self)
        subset.parent_mask = mask
        subset.x = self.x[mask]
        subset.y = self.y[mask]
        if self.unique_labels is not None:
            subset.unique_labels = self.unique_labels[mask]
            subset.unique_mapping = self.unique_mapping
        return subset

    def from_mask_optimized(self, mask):
        empty_object = type('', (), {})()
        empty_object.parent_mask = mask
        empty_object.get_single_labels = lambda: self.y[mask]
        return empty_object



from abc import ABC

from dtcontrol.decision_tree.determinization.non_determinizer import NonDeterminizer
from dtcontrol.decision_tree.impurity.impurity_measure import ImpurityMeasure

class DeterministicImpurityMeasure(ImpurityMeasure, ABC):

    def __init__(self, determinizer=NonDeterminizer()):
        self.determinizer = determinizer



import sys

import numpy as np

from dtcontrol.decision_tree.impurity.multi_label_impurity_measure import MultiLabelImpurityMeasure

class ScaledBincount(MultiLabelImpurityMeasure):
    def __init__(self, scaling_function):
        self.scaling_function = scaling_function

    def calculate_impurity(self, dataset, split):
        if len(split.get_masks(dataset)) == 1:
            return sys.maxsize
        y = dataset.get_single_labels()
        impurity = 0
        for mask in split.get_masks(dataset):
            subset = y[mask, :]
            if len(subset) == 0:
                return sys.maxsize
            impurity += (len(subset) / len(y)) * self.calculate_scaled_bincount(subset)
        return impurity

    def calculate_scaled_bincount(self, y):
        flattened_labels = y.flatten()
        flattened_labels = flattened_labels[flattened_labels != -1]  # -1 is only a filler
        label_counts = np.bincount(flattened_labels)[1:]
        if np.any(label_counts == len(y)):
            return 0
        label_counts = label_counts[label_counts != 0].astype('float')
        label_counts = label_counts / len(y)
        return len(label_counts) - sum(self.scaling_function(label_counts))



import numpy as np

from dtcontrol.decision_tree.determinization.non_determinizer import NonDeterminizer
from dtcontrol.decision_tree.splitting.linear_split import LinearSplit
from dtcontrol.decision_tree.splitting.splitting_strategy import SplittingStrategy

class LinearClassifierSplittingStrategy(SplittingStrategy):
    def __init__(self, classifier_class, determinizer=NonDeterminizer(), **kwargs):
        self.determinizer = determinizer
        self.classifier_class = classifier_class
        self.kwargs = kwargs

    def find_split(self, dataset, impurity_measure):
        x_numeric = dataset.get_numeric_x()
        if x_numeric.shape[1] == 0:
            return None

        y = self.determinizer.determinize(dataset)
        splits = {}
        for label in np.unique(y):
            new_y = np.copy(y)
            label_mask = (new_y == label)
            new_y[label_mask] = 1
            new_y[~label_mask] = -1
            classifier = self.classifier_class(**self.kwargs)
            classifier.fit(x_numeric, new_y)
            real_features = LinearSplit.map_numeric_coefficients_back(classifier.coef_[0], dataset)
            split = LinearClassifierSplit(classifier, real_features, dataset.numeric_columns)
            splits[split] = impurity_measure.calculate_impurity(dataset, split)

        return min(splits.keys(), key=splits.get)

class LinearClassifierSplit(LinearSplit):
    def __init__(self, classifier, real_coefficients, numeric_columns):
        super().__init__(classifier.coef_[0], classifier.intercept_[0], real_coefficients, numeric_columns)
        self.classifier = classifier
        self.numeric_columns = numeric_columns

    def get_masks(self, dataset):
        mask = self.classifier.predict(dataset.get_numeric_x()) == -1
        return [mask, ~mask]

    def predict(self, features):
        return 0 if self.classifier.predict(features[:, self.numeric_columns])[0] == -1 else 1



import os
import subprocess
import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

def git(*args):
    return subprocess.check_output(["git"] + list(args))

try:
    # Try to obtain version from the latest tag
    version = git("describe", "--tags").decode().strip()
except subprocess.CalledProcessError:
    # If not git repo or if tags are not available, use the version 'master'
    version = 'master'
# The VERSION file should not be manually edited, it is updated by the CI job
if os.path.exists(os.path.join(".", 'VERSION')):
    with open(os.path.join(".", 'VERSION')) as version_file:
        version = version_file.read().strip()

setuptools.setup(
    name="dtcontrol",
    version=version,
    description="A small tool which can convert automatically synthesised formally verified controllers into concise decision trees.",
    long_description_content_type="text/markdown",
    long_description=long_description,
    author='Mathias Jackermeier',
    author_email='mathias.jackermeier@outlook.de',
    license='MIT',
    url="https://dtcontrol.model.in.tum.de/",
    packages=['dtcontrol',
              'dtcontrol.c_templates',
              'dtcontrol.dataset',
              'dtcontrol.decision_tree',
              'dtcontrol.decision_tree.determinization',
              'dtcontrol.decision_tree.impurity',
              'dtcontrol.decision_tree.pre_processing',
              'dtcontrol.decision_tree.splitting',
              'dtcontrol.decision_tree.OC1_source',
              'dtcontrol.post_processing',
              'dtcontrol.ui',
              ],
    entry_points={
        'console_scripts': ['dtcontrol=dtcontrol.cli:main'],
    },
    classifiers=[
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        'License :: OSI Approved :: MIT License'
    ],
    python_requires='>=3.6',
    install_requires=[
        'Jinja2==2.10.3',
        'pandas==0.25.2',
        'psutil==5.6.7',
        'pydot==1.4.1',
        'ruamel.yaml==0.16.10',
        'scikit-learn==0.22',
        'tabulate==0.8.6',
        'tqdm==4.42.0'
    ],
    package_data={
        'dtcontrol': ['config.yml'],
        'dtcontrol.c_templates': ['*.c'],
        'dtcontrol.ui': ['*.js', '*.css', '*.html', '*.py'],
        'dtcontrol.decision_tree.OC1_source': ['*.c', '*.h', 'makefile', '*.readme', 'README'],
    }
)



import sys

import numpy as np

from dtcontrol.decision_tree.impurity.multi_label_impurity_measure import MultiLabelImpurityMeasure

class MultiLabelTwoingRule(MultiLabelImpurityMeasure):
    def calculate_impurity(self, dataset, split):
        if len(split.get_masks(dataset)) == 1:
            return sys.maxsize
        assert len(split.get_masks(dataset)) == 2
        y = dataset.get_single_labels()

        [left_mask, right_mask] = split.get_masks(dataset)
        left = y[left_mask]
        left_flat = left.flatten()
        left_flat = left_flat[left_flat != -1]
        right = y[right_mask]
        right_flat = right.flatten()
        right_flat = right_flat[right_flat != -1]
        if len(left) == 0 or len(right) == 0:
            return sys.maxsize
        num_labels = len(y)
        twoing_value = (len(left) / num_labels) * (len(right) / num_labels)

        left_index = -1
        left_counts = np.bincount(left_flat)[1:]
        r = np.where(left_counts == len(left))
        if len(r[0]) > 0:
            left_index = r[0][0] + 1

        right_index = -1
        right_counts = np.bincount(right_flat)[1:]
        r = np.where(right_counts == len(right))
        if len(r[0]) > 0:
            right_index = r[0][0] + 1

        s = 0
        y_flat = y.flatten()
        y_flat = y_flat[y_flat != -1]
        unique = np.unique(y_flat)
        for u in unique:
            num_left = len(left_flat[left_flat == u]) if left_index == -1 else (len(left) if u == left_index else 0)
            num_right = len(right_flat[right_flat == u]) if right_index == -1 else (len(right) if u == right_index else 0)
            s += abs(num_left / len(left) - num_right / len(right))
        twoing_value *= s * s
        if twoing_value == 0:
            return sys.maxsize
        return 1 / twoing_value



import sys

from dtcontrol.decision_tree.decision_tree import DecisionTree
from dtcontrol.post_processing.post_processing_method import PostProcessingMethod
from dtcontrol.util import make_set

class SafePruning(PostProcessingMethod):
    def __init__(self, classifier, rounds=None):
        if not isinstance(classifier, DecisionTree):  # in theory we also allow other classifiers
            raise ValueError('Safe pruning works only on decision trees.')
        self.name = classifier.get_name()
        super().__init__(classifier, self.name)
        self.rounds = rounds

    def run(self):
        self.prune(self.classifier.root, self.rounds)

    def prune(self, node, rounds):
        if node.is_leaf():
            return None if rounds is None else rounds
        remaining_rounds = None if rounds is None else sys.maxsize
        for child in node.children:
            r = self.prune(child, rounds)
            if rounds is not None and r < remaining_rounds:
                remaining_rounds = r
        if rounds is not None:
            assert remaining_rounds >= 0
        node.num_nodes = 1 + sum([c.num_nodes for c in node.children])
        node.num_inner_nodes = 1 + sum([c.num_inner_nodes for c in node.children])
        if remaining_rounds == 0 or any(not c.is_leaf() for c in node.children):
            return 0

        intersection = set.intersection(*[make_set(c.index_label) for c in node.children])
        if len(intersection) == 0:
            return None if rounds is None else remaining_rounds - 1
        node.index_label = list(intersection) if len(intersection) > 1 else list(intersection)[0]
        intersection = set.intersection(*[make_set(c.actual_label) for c in node.children])
        node.actual_label = list(intersection) if len(intersection) > 1 else list(intersection)[0]
        node.children = []
        node.num_nodes = 1
        node.num_inner_nodes = 0
        return None if rounds is None else remaining_rounds - 1



from abc import ABC, abstractmethod

from dtcontrol.dataset.multi_output_dataset import MultiOutputDataset
from dtcontrol.dataset.single_output_dataset import SingleOutputDataset

class LabelPreProcessor(ABC):

    def preprocess(self, dataset):
        """
        Returns the pre-processed dataset.
        """
        if isinstance(dataset, SingleOutputDataset):
            copy = SingleOutputDataset(dataset.filename)
            copy.copy_from_other_dataset(dataset)
            copy.y = self.preprocess_single_output(dataset)
        else:
            copy = MultiOutputDataset(dataset.filename)
            copy.copy_from_other_dataset(dataset)
            copy.y = self.preprocess_multi_output(dataset)
        return copy

    @abstractmethod
    def preprocess_single_output(self, dataset):
        pass

    @abstractmethod
    def preprocess_multi_output(self, dataset):
        pass



import os
import unittest
from unittest import SkipTest

from sklearn.linear_model import LogisticRegression

from dtcontrol.benchmark_suite import BenchmarkSuite
from dtcontrol.decision_tree.decision_tree import DecisionTree
from dtcontrol.decision_tree.determinization.max_freq_determinizer import MaxFreqDeterminizer
from dtcontrol.decision_tree.impurity.entropy import Entropy
from dtcontrol.decision_tree.impurity.multi_label_entropy import MultiLabelEntropy
from dtcontrol.decision_tree.pre_processing.norm_pre_processor import NormPreProcessor
from dtcontrol.decision_tree.splitting.axis_aligned import AxisAlignedSplittingStrategy
from dtcontrol.decision_tree.splitting.categorical_multi import CategoricalMultiSplittingStrategy
from dtcontrol.decision_tree.splitting.linear_classifier import LinearClassifierSplittingStrategy

class IntegrationTest(unittest.TestCase):
    def setUp(self):
        self.suite = BenchmarkSuite(timeout=60 * 60 * 2,
                                    save_folder='test_saved_classifiers',
                                    benchmark_file='benchmark_test',
                                    rerun=True)
        self.expected_results = {
            'cartpole': {
                'CART': 127,
                'logreg': 100,  # this is different from the table but apparently correct
                'OC1': 92,
                'maxfreq': 6,
                'maxfreq-logreg': 7,
                'minnorm': 56,
                'minnorm-logreg': 16,
                'multilabel': 4
            },
            'cruise-latest': {
                'CART': 494,
                'logreg': 392,
                'OC1': 290,
                'maxfreq': 2,
                'maxfreq-logreg': 2,
                'minnorm': 282,
                'minnorm-logreg': 197,
                'multilabel': 2
            },
            'dcdc': {
                'CART': 136,
                'logreg': 70,  # again different
                'maxfreq': 5,
                'maxfreq-logreg': 5,
                'minnorm': 11,
                'minnorm-logreg': 125,
                'multilabel': 3
            },
            '10rooms': {
                'CART': 8649,
                'logreg': 74,
                'OC1': 903,
                'maxfreq': 4,
                'maxfreq-logreg': 10,
                'minnorm': 2704,
                'minnorm-logreg': 28,
                'multilabel': 4
            },
            'vehicle': {
                'CART': 6619,
                'logreg': 5195,  # again different from table
                'OC1': 4639
            },
            'firewire_abst': {
                'CategoricalCART': 13,
                'AVG': 6
            },
            'wlan0': {
                'CategoricalCART': 135,
                'AVG': 106
            }
        }
        self.init_classifiers()
        if os.path.exists('benchmark_test.json'):
            os.remove('benchmark_test.json')

    def init_classifiers(self):
        self.cart = DecisionTree([AxisAlignedSplittingStrategy()], Entropy(), 'CART')
        self.maxfreq = DecisionTree([AxisAlignedSplittingStrategy()], Entropy(MaxFreqDeterminizer()), 'maxfreq',
                                    early_stopping=True, early_stopping_optimized=True)
        self.minnorm = DecisionTree([AxisAlignedSplittingStrategy()], Entropy(), 'minnorm',
                                    label_pre_processor=NormPreProcessor(min))
        logreg_strategy = LinearClassifierSplittingStrategy(LogisticRegression, solver='lbfgs', penalty='none')
        self.logreg = DecisionTree([AxisAlignedSplittingStrategy(), logreg_strategy], Entropy(), 'logreg')
        self.maxfreq_logreg = DecisionTree([AxisAlignedSplittingStrategy(), logreg_strategy],
                                           Entropy(MaxFreqDeterminizer()),
                                           'maxfreq-logreg', early_stopping=True, early_stopping_optimized=True)
        self.minnorm_logreg = DecisionTree([AxisAlignedSplittingStrategy(), logreg_strategy], Entropy(), 'minnorm-logreg',
                                           label_pre_processor=NormPreProcessor(min))
        self.multilabel = DecisionTree([AxisAlignedSplittingStrategy()], MultiLabelEntropy(), 'multilabel',
                                       early_stopping=True, early_stopping_optimized=True)
        self.categorical_cart = DecisionTree([AxisAlignedSplittingStrategy(), CategoricalMultiSplittingStrategy()],
                                             Entropy(), 'CategoricalCART')
        self.avg = DecisionTree([AxisAlignedSplittingStrategy(), CategoricalMultiSplittingStrategy(value_grouping=True)],
                                Entropy(), 'AVG')

    def test_fast(self):  # takes about 30s on my laptop
        datasets = ['cartpole', '10rooms', 'vehicle']
        classifiers = [self.cart, self.maxfreq, self.minnorm, self.multilabel]
        self.run_test(datasets, classifiers)

    def test_categorical(self):
        datasets = ['firewire_abst', 'wlan0']
        classifiers = [self.categorical_cart, self.avg]
        self.run_test(datasets, classifiers)

    @SkipTest
    def test_medium(self):  # takes about 4 min on my laptop
        datasets = ['cartpole', '10rooms', 'vehicle']
        classifiers = [self.cart, self.logreg, self.maxfreq, self.maxfreq_logreg, self.minnorm, self.multilabel]
        self.run_test(datasets, classifiers)

    @SkipTest
    def test_slow(self):  # takes about 6h on my laptop
        datasets = [
            'cartpole',
            'cruise-latest',
            'dcdc',
            '10rooms',
            'vehicle'
        ]
        classifiers = [self.cart, self.logreg, self.maxfreq, self.maxfreq_logreg, self.minnorm,
                       self.minnorm_logreg, self.multilabel]
        self.run_test(datasets, classifiers)

    def run_test(self, datasets, classifiers):
        # docker and local folders
        self.suite.add_datasets(['../examples', '../examples/prism', '/examples', '/examples/prism'], include=datasets)
        self.suite.benchmark(classifiers)
        self.assert_results_almost_equal(self.expected_results, self.suite.results)

    def assert_results_almost_equal(self, expected, actual, tol_percent=5):
        for ds in actual:
            for classifier in actual[ds]['classifiers']:
                if classifier not in expected[ds] or 'stats' not in actual[ds]['classifiers'][classifier]:
                    continue
                expected_paths = expected[ds][classifier]
                stats = actual[ds]['classifiers'][classifier]['stats']
                self.assertFalse('accuracy' in stats, 'Accuracy not 1.0')
                actual_paths = stats['paths']
                tol = (tol_percent / 100) * expected_paths
                tol = max(tol, 5)  # fix for small trees
                self.assertTrue(expected_paths - tol <= actual_paths <= expected_paths + tol,
                                f'Expected: [{expected_paths - tol}, {expected_paths + tol}], Actual: {actual_paths}'
                                f' with {classifier} on {ds}')

if __name__ == '__main__':
    unittest.main()



import logging
import re

import numpy as np
import pandas as pd
from tqdm import tqdm

from dtcontrol.dataset.dataset_loader import DatasetLoader

class UppaalDatasetLoader(DatasetLoader):
    def _load_dataset(self, filename):
        """
            Some assumptions regarding the Uppaal dataset
            1. All controllable states are named *.Choose
            2. Uppaal only works on integer domains (specifically np.int16)
            3. Actions are single-dimensional. However, if multiple components
               may have the .Choose state, they are added as categorical variables
               in x
            4. Step sizes for all state variables is 1
        """

        with open(filename) as f:
            logging.info("Reading from %s" % filename)

            action_set = set()
            controllable_state_set = set()

            num_lines = sum(1 for line in f)

            f.seek(0)

            logging.info("Extracting actions and controllable components from UPPPAL dump")
            for i, line in enumerate(tqdm(f, total=num_lines)):
                if i == 1:
                    # Extract numeric features
                    numeric_features = re.findall(r'(\w+)=-?[0-9]+', line)

                # Extract actions and controllable components
                if line.startswith('State: '):
                    ctrl = [word for word in line.split() if "Choose" in word]
                    if ctrl:
                        controllable_state_set.update(ctrl)
                if line.startswith('When'):
                    action_set.add(line[line.index(' take transition ') + 17:].rstrip())

            action_set = {action for action in action_set if "Choose" in action}
            actions = dict(zip(list(action_set), range(1, len(action_set) + 1)))
            controllable_states = list(controllable_state_set)

            # Figure out the assignments in each action and extract
            # the assigned value. The assigned variable is extracted
            # too, but not used anywhere as of now.
            index_to_actual = dict()
            action_var = None
            for (action, index) in actions.items():
                _, action_var, val = re.findall(r"((\w+) := (-?[0-9]+))", action)[0]
                index_to_actual[index] = int(val)

            row_num_vals = []
            row_actions = []

            ignore_current = False
            current_actions = []
            total_rows = 0
            total_state_actions = 0

            f.seek(0)

            for i in range(7):
                f.readline()

            logging.info("Extracting state-action pairs from UPPAAL dump")
            for i, line in enumerate(tqdm(f, total=num_lines)):
                if line.startswith("State:"):
                    # find if the state is controllable, and if so, then make that position 1 in categorical vals
                    controllable = False
                    cat_vals = [0 for i in range(len(controllable_states))]
                    for i in range(len(controllable_states)):
                        if controllable_states[i] in line:
                            cat_vals[i] = 1
                            controllable = True
                    if not controllable:
                        ignore_current = True
                        continue
                    else:
                        ignore_current = False
                        numeric_vals = [word.split("=")[1] for word in line.split() if "=" in word]
                elif ignore_current:
                    continue
                elif line.startswith("When"):
                    action_str = line[line.index(' take transition ') + 17:].rstrip()
                    current_actions.append(actions[action_str])
                elif line.startswith("While"):
                    # We implicityly assume that transitions starting with 'While' are mapped to wait.
                    ignore_current = True
                elif line.strip() == "":
                    if not ignore_current:
                        row_num_vals.append(cat_vals + numeric_vals)
                        row_actions.append(current_actions)
                        total_rows += 1
                        total_state_actions += len(current_actions)
                    ignore_current = False
                    current_actions = []
                else:
                    raise Exception("ERROR: Unhandled line in input")

            logging.info(
                f"Done reading {total_rows} states with \na total of {total_state_actions} state-action pairs.")

            # Project onto measurable variables, the strategy should not depend on the gua variables coming from euler
            projection_variables = controllable_states + list(filter(lambda x: 'gua' not in x, numeric_features))
            num_df = pd.DataFrame(row_num_vals, columns=controllable_states + numeric_features, dtype=np.int16)
            num_df = num_df[projection_variables]

            grouped = num_df.groupby(projection_variables)
            x = np.empty((len(grouped), len(projection_variables)), dtype=np.int16)
            y = np.full((len(grouped), max([len(y) for y in row_actions])), -1, dtype=np.int16)

            i = 0
            for (group, indices) in grouped.indices.items():
                if len(indices) > 1:
                    x[i] = group
                    conservative_actions = set(actions.values()).copy()
                    for idx in indices:
                        conservative_actions &= set(row_actions[idx])
                    assert len(
                        conservative_actions) > 0, "Stategy for picking safe action doesn't work. Deeper analysis needed."
                    y[i][0:len(conservative_actions)] = list(sorted(conservative_actions))
                else:
                    x[i] = group
                    y[i][0:len(row_actions[indices[0]])] = sorted(row_actions[indices[0]])
                i = i + 1

            logging.info("Constructed training set with %s datapoints" % x.shape[0])

            # construct metadata
            # assumption is that UPPAAL only works with integers
            x_metadata = dict()
            x_metadata["variables"] = projection_variables
            x_metadata["categorical"] = []
            x_metadata["min"] = [int(i) for i in np.amin(x, axis=0)]
            x_metadata["max"] = [int(i) for i in np.amax(x, axis=0)]
            x_metadata["step_size"] = [1 for _ in range(len(projection_variables))]

            y_metadata = dict()
            y_metadata["variables"] = [action_var]
            y_metadata["min"] = [min(index_to_actual.values())]
            y_metadata["max"] = [max(index_to_actual.values())]
            y_metadata["step_size"] = [int((y_metadata["max"][0] - y_metadata["min"][0]) / (len(index_to_actual) - 1))]

            logging.debug(x_metadata)
            logging.debug(y_metadata)

            return (x, x_metadata, y, y_metadata, index_to_actual)
